/*----- PROTECTED REGION ID(CATSync.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        CATSync.cpp
//
// description : C++ source for the CATSync class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               CATSync are implemented in this file.
//
// project :     CATSync
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "CATSync.h"
#include "CATSyncClass.h"

#include "catsyncthread.h"
#include "matchthread.h"
#include "sequencerthread.h"

#ifdef DPDK
	#include <gof2_bpm_corr_tbl.h>
	#include <gof2_fun.c>
#endif

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync.cpp

/**
 *  CATSync class description:
 *    Tango server used to synchronize CAT simulator with field
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name             |  Method name
//================================================================
//  State                    |  Inherited (no method)
//  Status                   |  Inherited (no method)
//  Start                    |  start
//  Stop                     |  stop
//  GotoEnergy               |  goto_energy
//  UpdateEnergy             |  update_energy
//  SyncSourcePointsToHLS    |  sync_source_points_to_hls
//  SwitchLFBCAT             |  switch_lfbcat
//  SwitchLFBHLS             |  switch_lfbhls
//  ConvertSourcePointsMM    |  convert_source_points_mm
//  ResetSourcePointOffsets  |  reset_source_point_offsets
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  EnableCATEnergyUpdate       |  Tango::DevBoolean	Scalar
//  EnableCATSourcePointUpdate  |  Tango::DevBoolean	Scalar
//  LFBCATTimeFactor            |  Tango::DevDouble	Scalar
//  LFBHLSTimeFactor            |  Tango::DevDouble	Scalar
//  EnableCATMatchOrbit         |  Tango::DevBoolean	Scalar
//  EnableCATSyncMagnets        |  Tango::DevBoolean	Scalar
//  CATState                    |  Tango::DevState	Scalar
//  HLSState                    |  Tango::DevState	Scalar
//  SyncGain                    |  Tango::DevDouble	Scalar
//  CATTuneHor                  |  Tango::DevDouble	Scalar
//  CATTuneVer                  |  Tango::DevDouble	Scalar
//  MatchThreadErrMsg           |  Tango::DevString	Spectrum  ( max = 10000)
//  SyncThreadErrMsg            |  Tango::DevString	Spectrum  ( max = 10000)
//  MatchThreadErrCnt           |  Tango::DevULong	Spectrum  ( max = 10000)
//  SyncThreadErrCnt            |  Tango::DevULong	Spectrum  ( max = 10000)
//  MatchThreadCnt              |  Tango::DevULong	Spectrum  ( max = 10000)
//  SyncThreadCnt               |  Tango::DevULong	Spectrum  ( max = 10000)
//================================================================

namespace CATSync_ns
{
/*----- PROTECTED REGION ID(CATSync::namespace_starting) ENABLED START -----*/
/* clang-format on */
//	static initializations
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : CATSync::CATSync()
 * Description:  Constructors for a Tango device
 *                implementing the classCATSync
 */
//--------------------------------------------------------
CATSync::CATSync(Tango::DeviceClass *cl, std::string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(CATSync::constructor_1) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::constructor_1
}
//--------------------------------------------------------
CATSync::CATSync(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(CATSync::constructor_2) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::constructor_2
}
//--------------------------------------------------------
CATSync::CATSync(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(CATSync::constructor_3) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::constructor_3
}
//--------------------------------------------------------
CATSync::~CATSync()
{
	delete_device();
}

//--------------------------------------------------------
/**
 *	Method      : CATSync::delete_device()
 * Description:  will be called at device destruction or at init command
 */
//--------------------------------------------------------
void CATSync::delete_device()
{
	DEBUG_STREAM << "CATSync::delete_device() " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::delete_device) ENABLED START -----*/
	/* clang-format on */
	//	Delete device allocated objects
	stop();
	cleanup_devices();

	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		delete attr_SyncThreadErrMsg_read[i];
	}

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
		delete attr_MatchThreadErrMsg_read[i];
	}	

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::delete_device
	delete[] attr_EnableCATEnergyUpdate_read;
	delete[] attr_EnableCATSourcePointUpdate_read;
	delete[] attr_LFBCATTimeFactor_read;
	delete[] attr_LFBHLSTimeFactor_read;
	delete[] attr_EnableCATMatchOrbit_read;
	delete[] attr_EnableCATSyncMagnets_read;
	delete[] attr_CATState_read;
	delete[] attr_HLSState_read;
	delete[] attr_SyncGain_read;
	delete[] attr_CATTuneHor_read;
	delete[] attr_CATTuneVer_read;
	delete[] attr_MatchThreadErrMsg_read;
	delete[] attr_SyncThreadErrMsg_read;
	delete[] attr_MatchThreadErrCnt_read;
	delete[] attr_SyncThreadErrCnt_read;
	delete[] attr_MatchThreadCnt_read;
	delete[] attr_SyncThreadCnt_read;
}

//--------------------------------------------------------
/**
 *	Method      : CATSync::init_device()
 * Description:  will be called at device initialization.
 */
//--------------------------------------------------------
void CATSync::init_device()
{
	DEBUG_STREAM << "CATSync::init_device() create device " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::init_device_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialization before get_device_property() call
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::init_device_before


	//	Get the device properties from database
	get_device_property();

	attr_EnableCATEnergyUpdate_read = new Tango::DevBoolean[1];
	attr_EnableCATSourcePointUpdate_read = new Tango::DevBoolean[1];
	attr_LFBCATTimeFactor_read = new Tango::DevDouble[1];
	attr_LFBHLSTimeFactor_read = new Tango::DevDouble[1];
	attr_EnableCATMatchOrbit_read = new Tango::DevBoolean[1];
	attr_EnableCATSyncMagnets_read = new Tango::DevBoolean[1];
	attr_CATState_read = new Tango::DevState[1];
	attr_HLSState_read = new Tango::DevState[1];
	attr_SyncGain_read = new Tango::DevDouble[1];
	attr_CATTuneHor_read = new Tango::DevDouble[1];
	attr_CATTuneVer_read = new Tango::DevDouble[1];
	attr_MatchThreadErrMsg_read = new Tango::DevString[10000];
	attr_SyncThreadErrMsg_read = new Tango::DevString[10000];
	attr_MatchThreadErrCnt_read = new Tango::DevULong[10000];
	attr_SyncThreadErrCnt_read = new Tango::DevULong[10000];
	attr_MatchThreadCnt_read = new Tango::DevULong[10000];
	attr_SyncThreadCnt_read = new Tango::DevULong[10000];
	/*----- PROTECTED REGION ID(CATSync::init_device) ENABLED START -----*/
	/* clang-format on */
	//	Initialize device
	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		attr_SyncThreadErrCnt_read[i] = 0;
		attr_SyncThreadCnt_read[i] = 0;
		attr_SyncThreadErrMsg_read[i] = new char [CATSYNC_LEN];
		memset(attr_SyncThreadErrMsg_read[i], 0, CATSYNC_LEN);
	}

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
		attr_MatchThreadErrCnt_read[i] = 0;
		attr_MatchThreadCnt_read[i] = 0;
		attr_MatchThreadErrMsg_read[i] = new char [CATSYNC_LEN];
		memset(attr_MatchThreadErrMsg_read[i], 0, CATSYNC_LEN);
	}
	*attr_EnableCATEnergyUpdate_read = false;
	*attr_EnableCATSourcePointUpdate_read = false;
	*attr_EnableCATMatchOrbit_read = false;
	*attr_EnableCATSyncMagnets_read = false;

	*attr_LFBCATTimeFactor_read = 1;
	*attr_LFBHLSTimeFactor_read = 1;
	*attr_SyncGain_read = 1;

	*attr_CATState_read = Tango::UNKNOWN;
	*attr_HLSState_read = Tango::UNKNOWN;

#ifdef DPDK
	if (hugePages)
		enable_hugepages = 1;
	else
		enable_hugepages = 0;


	if (enable_hugepages)
		client_locate_shm(GOF2_SHM_HUGE_ID, &gof2, enable_hugepages);	
	else
		client_locate_shm(GOF2_SHM_ID, &gof2, enable_hugepages);
#endif
	connect_devices();

	if (autoStart)
		start();
	

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::init_device
}

//--------------------------------------------------------
/**
 *	Method      : CATSync::get_device_property()
 * Description:  Read database to initialize property data members.
 */
//--------------------------------------------------------
void CATSync::get_device_property()
{
	/*----- PROTECTED REGION ID(CATSync::get_device_property_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialize property data members
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("MagnetConfiguration"));
	dev_prop.push_back(Tango::DbDatum("PreSequencer"));
	dev_prop.push_back(Tango::DbDatum("PostSequencer"));
	dev_prop.push_back(Tango::DbDatum("AutoStart"));
	dev_prop.push_back(Tango::DbDatum("CATDSDevice"));
	dev_prop.push_back(Tango::DbDatum("HugePages"));
	dev_prop.push_back(Tango::DbDatum("SharedMemoryID"));
	dev_prop.push_back(Tango::DbDatum("EnergyDevice"));
	dev_prop.push_back(Tango::DbDatum("EnergyRanges"));
	dev_prop.push_back(Tango::DbDatum("HLSDevice"));
	dev_prop.push_back(Tango::DbDatum("SourcePointLabels"));
	dev_prop.push_back(Tango::DbDatum("EnergyLoopPeriodSec"));
	dev_prop.push_back(Tango::DbDatum("SyncLoopPeriodSec"));
	dev_prop.push_back(Tango::DbDatum("MIMOFBList"));
	dev_prop.push_back(Tango::DbDatum("PsDelayedStartList"));
	dev_prop.push_back(Tango::DbDatum("OrbitDelayedStartList"));
	dev_prop.push_back(Tango::DbDatum("TuneHorDevice"));
	dev_prop.push_back(Tango::DbDatum("TuneVerDevice"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);

		//	get instance on CATSyncClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		CATSyncClass	*ds_class =
			(static_cast<CATSyncClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize MagnetConfiguration from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  magnetConfiguration;
		else {
			//	Try to initialize MagnetConfiguration from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  magnetConfiguration;
		}
		//	And try to extract MagnetConfiguration value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  magnetConfiguration;

		//	Try to initialize PreSequencer from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  preSequencer;
		else {
			//	Try to initialize PreSequencer from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  preSequencer;
		}
		//	And try to extract PreSequencer value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  preSequencer;

		//	Try to initialize PostSequencer from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  postSequencer;
		else {
			//	Try to initialize PostSequencer from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  postSequencer;
		}
		//	And try to extract PostSequencer value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  postSequencer;

		//	Try to initialize AutoStart from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoStart;
		else {
			//	Try to initialize AutoStart from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoStart;
		}
		//	And try to extract AutoStart value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoStart;

		//	Try to initialize CATDSDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  cATDSDevice;
		else {
			//	Try to initialize CATDSDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  cATDSDevice;
		}
		//	And try to extract CATDSDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cATDSDevice;

		//	Try to initialize HugePages from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  hugePages;
		else {
			//	Try to initialize HugePages from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  hugePages;
		}
		//	And try to extract HugePages value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  hugePages;

		//	Try to initialize SharedMemoryID from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  sharedMemoryID;
		else {
			//	Try to initialize SharedMemoryID from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  sharedMemoryID;
		}
		//	And try to extract SharedMemoryID value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  sharedMemoryID;

		//	Try to initialize EnergyDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  energyDevice;
		else {
			//	Try to initialize EnergyDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  energyDevice;
		}
		//	And try to extract EnergyDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  energyDevice;

		//	Try to initialize EnergyRanges from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  energyRanges;
		else {
			//	Try to initialize EnergyRanges from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  energyRanges;
		}
		//	And try to extract EnergyRanges value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  energyRanges;

		//	Try to initialize HLSDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  hLSDevice;
		else {
			//	Try to initialize HLSDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  hLSDevice;
		}
		//	And try to extract HLSDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  hLSDevice;

		//	Try to initialize SourcePointLabels from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  sourcePointLabels;
		else {
			//	Try to initialize SourcePointLabels from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  sourcePointLabels;
		}
		//	And try to extract SourcePointLabels value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  sourcePointLabels;

		//	Try to initialize EnergyLoopPeriodSec from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  energyLoopPeriodSec;
		else {
			//	Try to initialize EnergyLoopPeriodSec from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  energyLoopPeriodSec;
		}
		//	And try to extract EnergyLoopPeriodSec value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  energyLoopPeriodSec;

		//	Try to initialize SyncLoopPeriodSec from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  syncLoopPeriodSec;
		else {
			//	Try to initialize SyncLoopPeriodSec from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  syncLoopPeriodSec;
		}
		//	And try to extract SyncLoopPeriodSec value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  syncLoopPeriodSec;

		//	Try to initialize MIMOFBList from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  mIMOFBList;
		else {
			//	Try to initialize MIMOFBList from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  mIMOFBList;
		}
		//	And try to extract MIMOFBList value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  mIMOFBList;

		//	Try to initialize PsDelayedStartList from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  psDelayedStartList;
		else {
			//	Try to initialize PsDelayedStartList from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  psDelayedStartList;
		}
		//	And try to extract PsDelayedStartList value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  psDelayedStartList;

		//	Try to initialize OrbitDelayedStartList from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  orbitDelayedStartList;
		else {
			//	Try to initialize OrbitDelayedStartList from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  orbitDelayedStartList;
		}
		//	And try to extract OrbitDelayedStartList value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  orbitDelayedStartList;

		//	Try to initialize TuneHorDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  tuneHorDevice;
		else {
			//	Try to initialize TuneHorDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  tuneHorDevice;
		}
		//	And try to extract TuneHorDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  tuneHorDevice;

		//	Try to initialize TuneVerDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  tuneVerDevice;
		else {
			//	Try to initialize TuneVerDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  tuneVerDevice;
		}
		//	And try to extract TuneVerDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  tuneVerDevice;

	}

	/*----- PROTECTED REGION ID(CATSync::get_device_property_after) ENABLED START -----*/
	/* clang-format on */
	//	Check device property data members init
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : CATSync::always_executed_hook()
 * Description:  method always executed before any command is executed
 */
//--------------------------------------------------------
void CATSync::always_executed_hook()
{
	DEBUG_STREAM << "CATSync::always_executed_hook()  " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::always_executed_hook) ENABLED START -----*/
	/* clang-format on */
	//	code always executed before all requests
	if (run_catsync_flag[0]) {
		set_state(Tango::ON);set_status("Synchronization loop running");
	}
	else {
		set_state(Tango::ON);set_status("Synchronization loop stopped");
	}

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : CATSync::read_attr_hardware()
 * Description:  Hardware acquisition for attributes
 */
//--------------------------------------------------------
void CATSync::read_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "CATSync::read_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : CATSync::write_attr_hardware()
 * Description:  Hardware writing for attributes
 */
//--------------------------------------------------------
void CATSync::write_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "CATSync::write_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::write_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute EnableCATEnergyUpdate related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_EnableCATEnergyUpdate(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_EnableCATEnergyUpdate(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_EnableCATEnergyUpdate) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_EnableCATEnergyUpdate_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_EnableCATEnergyUpdate
}
//--------------------------------------------------------
/**
 *	Write attribute EnableCATEnergyUpdate related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_EnableCATEnergyUpdate(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_EnableCATEnergyUpdate(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_EnableCATEnergyUpdate) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_EnableCATEnergyUpdate_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_EnableCATEnergyUpdate
}
//--------------------------------------------------------
/**
 *	Read attribute EnableCATSourcePointUpdate related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_EnableCATSourcePointUpdate(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_EnableCATSourcePointUpdate(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_EnableCATSourcePointUpdate) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_EnableCATSourcePointUpdate_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_EnableCATSourcePointUpdate
}
//--------------------------------------------------------
/**
 *	Write attribute EnableCATSourcePointUpdate related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_EnableCATSourcePointUpdate(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_EnableCATSourcePointUpdate(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_EnableCATSourcePointUpdate) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_EnableCATSourcePointUpdate_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_EnableCATSourcePointUpdate
}
//--------------------------------------------------------
/**
 *	Read attribute LFBCATTimeFactor related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_LFBCATTimeFactor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_LFBCATTimeFactor(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_LFBCATTimeFactor) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_LFBCATTimeFactor_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_LFBCATTimeFactor
}
//--------------------------------------------------------
/**
 *	Write attribute LFBCATTimeFactor related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_LFBCATTimeFactor(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_LFBCATTimeFactor(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_LFBCATTimeFactor) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_LFBCATTimeFactor_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_LFBCATTimeFactor
}
//--------------------------------------------------------
/**
 *	Read attribute LFBHLSTimeFactor related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_LFBHLSTimeFactor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_LFBHLSTimeFactor(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_LFBHLSTimeFactor) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_LFBHLSTimeFactor_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_LFBHLSTimeFactor
}
//--------------------------------------------------------
/**
 *	Write attribute LFBHLSTimeFactor related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_LFBHLSTimeFactor(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_LFBHLSTimeFactor(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_LFBHLSTimeFactor) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_LFBHLSTimeFactor_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_LFBHLSTimeFactor
}
//--------------------------------------------------------
/**
 *	Read attribute EnableCATMatchOrbit related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_EnableCATMatchOrbit(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_EnableCATMatchOrbit(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_EnableCATMatchOrbit) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_EnableCATMatchOrbit_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_EnableCATMatchOrbit
}
//--------------------------------------------------------
/**
 *	Write attribute EnableCATMatchOrbit related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_EnableCATMatchOrbit(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_EnableCATMatchOrbit(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_EnableCATMatchOrbit) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_EnableCATMatchOrbit_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_EnableCATMatchOrbit
}
//--------------------------------------------------------
/**
 *	Read attribute EnableCATSyncMagnets related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_EnableCATSyncMagnets(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_EnableCATSyncMagnets(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_EnableCATSyncMagnets) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_EnableCATSyncMagnets_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_EnableCATSyncMagnets
}
//--------------------------------------------------------
/**
 *	Write attribute EnableCATSyncMagnets related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_EnableCATSyncMagnets(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_EnableCATSyncMagnets(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_EnableCATSyncMagnets) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_EnableCATSyncMagnets_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_EnableCATSyncMagnets
}
//--------------------------------------------------------
/**
 *	Read attribute CATState related method
 *
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_CATState(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_CATState(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_CATState) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	catdevice->read_attribute("State") >> *attr_CATState_read;
	attr.set_value(attr_CATState_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_CATState
}
//--------------------------------------------------------
/**
 *	Read attribute HLSState related method
 *
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_HLSState(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_HLSState(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_HLSState) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	hlsdevice->read_attribute("State") >> *attr_HLSState_read;
	attr.set_value(attr_HLSState_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_HLSState
}
//--------------------------------------------------------
/**
 *	Read attribute SyncGain related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_SyncGain(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_SyncGain(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_SyncGain) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_SyncGain_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_SyncGain
}
//--------------------------------------------------------
/**
 *	Write attribute SyncGain related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::write_SyncGain(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CATSync::write_SyncGain(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CATSync::write_SyncGain) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_SyncGain_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::write_SyncGain
}
//--------------------------------------------------------
/**
 *	Read attribute CATTuneHor related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_CATTuneHor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_CATTuneHor(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_CATTuneHor) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	catdevice->read_attribute("tune_x") >> *attr_CATTuneHor_read;
	attr.set_value(attr_CATTuneHor_read);

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_CATTuneHor
}
//--------------------------------------------------------
/**
 *	Read attribute CATTuneVer related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CATSync::read_CATTuneVer(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_CATTuneVer(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_CATTuneVer) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	catdevice->read_attribute("tune_y") >> *attr_CATTuneVer_read;
	attr.set_value(attr_CATTuneVer_read);

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_CATTuneVer
}
//--------------------------------------------------------
/**
 *	Read attribute MatchThreadErrMsg related method
 *
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void CATSync::read_MatchThreadErrMsg(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_MatchThreadErrMsg(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_MatchThreadErrMsg) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_MatchThreadErrMsg_read, CATSYNC_MATCH_ORB_NUM);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_MatchThreadErrMsg
}
//--------------------------------------------------------
/**
 *	Read attribute SyncThreadErrMsg related method
 *
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void CATSync::read_SyncThreadErrMsg(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_SyncThreadErrMsg(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_SyncThreadErrMsg) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_SyncThreadErrMsg_read, CATSYNC_MAX_THREADS);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_SyncThreadErrMsg
}
//--------------------------------------------------------
/**
 *	Read attribute MatchThreadErrCnt related method
 *
 *
 *	Data type:	Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void CATSync::read_MatchThreadErrCnt(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_MatchThreadErrCnt(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_MatchThreadErrCnt) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_MatchThreadErrCnt_read, CATSYNC_MATCH_ORB_NUM);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_MatchThreadErrCnt
}
//--------------------------------------------------------
/**
 *	Read attribute SyncThreadErrCnt related method
 *
 *
 *	Data type:	Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void CATSync::read_SyncThreadErrCnt(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_SyncThreadErrCnt(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_SyncThreadErrCnt) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_SyncThreadErrCnt_read, CATSYNC_MAX_THREADS);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_SyncThreadErrCnt
}
//--------------------------------------------------------
/**
 *	Read attribute MatchThreadCnt related method
 *
 *
 *	Data type:	Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void CATSync::read_MatchThreadCnt(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_MatchThreadCnt(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_MatchThreadCnt) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_MatchThreadCnt_read, CATSYNC_MATCH_ORB_NUM);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_MatchThreadCnt
}
//--------------------------------------------------------
/**
 *	Read attribute SyncThreadCnt related method
 *
 *
 *	Data type:	Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void CATSync::read_SyncThreadCnt(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CATSync::read_SyncThreadCnt(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(CATSync::read_SyncThreadCnt) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_SyncThreadCnt_read, CATSYNC_MAX_THREADS);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::read_SyncThreadCnt
}

//--------------------------------------------------------
/**
 *	Method      : CATSync::add_dynamic_attributes()
 * Description:  Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void CATSync::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(CATSync::add_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic attributes if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command Start related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::start()
{
	DEBUG_STREAM << "CATSync::Start()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::start) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	start_sequencer(CATSYNC_PRESEQUENCER_IDX);

	clock_gettime(CLOCK_REALTIME, &loop_start_time);

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
		orbit_start_time[i].tv_sec = loop_start_time.tv_sec;
		orbit_start_time[i].tv_nsec = loop_start_time.tv_nsec;
		orbit_start_time[i].tv_sec += orbitDelayedStartList[i]; // add delay
	}

	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		for (int j = 0; j < ps_magnet_sync[i].magnetdevice.size(); j++) {
			ps_magnet_sync[i].start_time[j].tv_sec = loop_start_time.tv_sec;
			ps_magnet_sync[i].start_time[j].tv_nsec = loop_start_time.tv_nsec;
			ps_magnet_sync[i].start_time[j].tv_sec += ps_magnet_sync[i].delay_start[j].tv_sec;
		}
	}


	// start the threads that sync magnets and simulator attributes
	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		if (run_catsync_flag[i] == false) {
			abort_catsync_flag[i] = false;
			err_catsync_flag[i] = false;	
			catsyncloop[i] = new catsyncthread(this,i);		
			catsyncloop[i]->start();
		}
	}
	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
		if (run_match_orb_flag[i] == false) {
			abort_match_orb_flag[i] = false;
			err_match_orb_flag[i] = false;	
			matchloop[i] = new matchthread(this,i);		
			matchloop[i]->start();
		}
	}

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::start
}
//--------------------------------------------------------
/**
 *	Command Stop related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::stop()
{
	DEBUG_STREAM << "CATSync::Stop()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::stop) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) 
		abort_catsync_flag[i] = true;

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) 
		abort_match_orb_flag[i] = true;

	for (int i = 0; i < CATSYNC_SEQUENCER_NUM; i++) 
		abort_sequencer_flag[i] = true;

	while (1) {
		int cnt = 0;
		for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
			if (run_catsync_flag[i] == true) {
				cnt++;
				break;
			}
		}
	
		for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
			if (run_match_orb_flag[i] == true) {
				cnt++;
				break;
			}
		}

		for (int i = 0; i < CATSYNC_SEQUENCER_NUM; i++) {
			if (run_sequencer_flag[i] == true) {
				cnt++;
				break;
			}
		}


		if (cnt == 0)
			break;
		else
			omni_thread::sleep(0, 100000000);
	}

	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) 
		abort_catsync_flag[i] = false;

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) 
		abort_match_orb_flag[i] = false;

	for (int i = 0; i < CATSYNC_SEQUENCER_NUM; i++) 
		abort_sequencer_flag[i] = false;

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::stop
}
//--------------------------------------------------------
/**
 *	Command GotoEnergy related method
 *
 *
 *	@param argin Energy (GeV)
 */
//--------------------------------------------------------
void CATSync::goto_energy(Tango::DevDouble argin)
{
	DEBUG_STREAM << "CATSync::GotoEnergy()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::goto_energy) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	// best effort, no error return
	for (int j = 0; j < CATSYNC_MAX_THREADS; j++) {
		device_set_energy(argin, j);
	}


	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::goto_energy
}
//--------------------------------------------------------
/**
 *	Command UpdateEnergy related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::update_energy()
{
	DEBUG_STREAM << "CATSync::UpdateEnergy()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::update_energy) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	double energy;
	energydevice->read_attribute(energy_attribute.c_str()) >> energy;

	/*
	srand((unsigned int)time(NULL));
	double random_number = (double)rand() / RAND_MAX;	
	energy = 2 + random_number * 0.001;
	*/
	
	for (int j = 0; j < CATSYNC_MAX_THREADS; j++) {
		for (int i = 0; i < min_energies.size(); i++) {
			if ((energy >= min_energies[i]) && (energy <= max_energies[i])) {
				goto_energy(energy);
				break;
			}
		}
		old_energy[j] = energy;
	}
	
     /*
     try {
        // Nome del dispositivo e dell'attributo
        std::string deviceName = "sr/diagnostics/bpmid";
        std::string attributeName = "X_B6";

        std::string fullAttributeName = deviceName + "/" + attributeName;
        Tango::AttributeProxy attrProxy(fullAttributeName);

               // Lettura delle proprietà dell'attributo
        Tango::DbData propertyList;
        propertyList.push_back(Tango::DbDatum("max_alarm"));
        propertyList.push_back(Tango::DbDatum("min_alarm"));
        propertyList.push_back(Tango::DbDatum("max_warning"));
        propertyList.push_back(Tango::DbDatum("min_warning"));

        attrProxy.get_property(propertyList);

        // Estrazione dei valori dalle proprietà
        double alarmHigh = 0.0, alarmLow = 0.0, warningHigh = 0.0, warningLow = 0.0;
        for (auto& prop : propertyList) {
            if (prop.name == "max_alarm") prop >> alarmHigh;
            else if (prop.name == "min_alarm") prop >> alarmLow;
            else if (prop.name == "max_warning") prop >> warningHigh;
            else if (prop.name == "min_warning") prop >> warningLow;
        }

        // Stampa dei risultati
        std::cout << "Attribute: " << attributeName << std::endl;
        std::cout << "Max Alarm: " << alarmHigh << std::endl;
        std::cout << "Min Alarm: " << alarmLow << std::endl;
        std::cout << "Max Warning: " << warningHigh << std::endl;
        std::cout << "Min Warning: " << warningLow << std::endl;

    } catch (Tango::DevFailed& e) {
        // Gestione degli errori specifici di Tango
        std::cerr << "Tango exception: " << e.errors[0].desc << std::endl;
    } catch (std::exception& e) {
        // Gestione di altre eccezioni
        std::cerr << "Standard exception: " << e.what() << std::endl;
    }
*/






	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::update_energy
}
//--------------------------------------------------------
/**
 *	Command SyncSourcePointsToHLS related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::sync_source_points_to_hls()
{
	DEBUG_STREAM << "CATSync::SyncSourcePointsToHLS()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::sync_source_points_to_hls) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	synchronize_attributes();

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::sync_source_points_to_hls
}
//--------------------------------------------------------
/**
 *	Command SwitchLFBCAT related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::switch_lfbcat()
{
	DEBUG_STREAM << "CATSync::SwitchLFBCAT()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::switch_lfbcat) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	updateSensorConfiguration(false);

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::switch_lfbcat
}
//--------------------------------------------------------
/**
 *	Command SwitchLFBHLS related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::switch_lfbhls()
{
	DEBUG_STREAM << "CATSync::SwitchLFBHLS()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::switch_lfbhls) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	updateSensorConfiguration(true);

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::switch_lfbhls
}
//--------------------------------------------------------
/**
 *	Command ConvertSourcePointsMM related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::convert_source_points_mm()
{
	DEBUG_STREAM << "CATSync::ConvertSourcePointsMM()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::convert_source_points_mm) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	convert_sourcepoints_mm();

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::convert_source_points_mm
}
//--------------------------------------------------------
/**
 *	Command ResetSourcePointOffsets related method
 *
 *
 */
//--------------------------------------------------------
void CATSync::reset_source_point_offsets()
{
	DEBUG_STREAM << "CATSync::ResetSourcePointOffsets()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(CATSync::reset_source_point_offsets) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	reset_sourcepoint_offsets();

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::reset_source_point_offsets
}
//--------------------------------------------------------
/**
 *	Method      : CATSync::add_dynamic_commands()
 * Description:  Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void CATSync::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(CATSync::add_dynamic_commands) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic commands if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSync::add_dynamic_commands
}

/*----- PROTECTED REGION ID(CATSync::namespace_ending) ENABLED START -----*/
/* clang-format on */
//	Additional Methods


//--------------------------------------------------------
/**
 *	Method      : CATSync::connect_devices()
 * Description:  Connect to tango devices
 *                for specified device.
 */
//--------------------------------------------------------
void CATSync::connect_devices()
{

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
		abort_match_orb_flag[i] = false;
		run_match_orb_flag[i] = false;
		err_match_orb_flag[i] = false;
	}

	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		abort_catsync_flag[i] = false;
		run_catsync_flag[i] = false;
		err_catsync_flag[i] = false;
	}

	for (int i = 0; i < CATSYNC_SEQUENCER_NUM; i++) {
		abort_sequencer_flag[i] = false;
		run_sequencer_flag[i] = false;
		err_sequencer_flag[i] = false;
	}

	
	clock_gettime(CLOCK_REALTIME, &source_points_update_time);
	source_points_update_time.tv_sec -= syncLoopPeriodSec;	

	// init data structures
	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		old_energy[i] = -1;
		clock_gettime(CLOCK_REALTIME, &energy_update_time[i]);
		energy_update_time[i].tv_sec -= energyLoopPeriodSec;
	}

	// define energy valid segments to update simulator and magnet energies
	for (int i = 0; i < energyRanges.size(); i++) {
		vector<string> res2;
		string_explode(energyRanges[i], ";", &res2);
		if (res2.size() == 2) {
			min_energies.push_back(atof(res2[0].c_str()));
			max_energies.push_back(atof(res2[1].c_str()));
		}
	}

	// define the list of all the sourcepoints
	for (int i = 0; i < sourcePointLabels.size(); i++) {
		if (sourcePointLabels[i][0] == '#')
			continue;
		vector<string> res2;
		string_explode(sourcePointLabels[i], ";", &res2);
		if (res2.size() > 0) {
			for (int j = 0; j < res2.size(); j++) {
				source_point_attributes.push_back(res2[j]);
			}
		}
	}

	// connect to simulator
	catdevice = 0;
	catdevice = new Tango::DeviceProxy(cATDSDevice);

	// connect to simulator
	hlsdevice = 0;
	hlsdevice = new Tango::DeviceProxy(hLSDevice);

	// connect to energy device
	energydevice = 0;
	vector<string> res1;
	string_explode(energyDevice, ";", &res1);
	if (res1.size() == 2) {
		energydevice = new Tango::DeviceProxy(res1[0]);
		energy_attribute = res1[1];
	}

	tunehor_device = 0;
	vector<string> resh;	
	string_explode(tuneHorDevice, ";", &resh);
	if (resh.size() == 2) {
		tunehor_device = new Tango::DeviceProxy(resh[0]);
		tunehor_attribute = resh[1];
	}	

	tunever_device = 0;
	vector<string> resv;	
	string_explode(tuneVerDevice, ";", &resv);
	if (resv.size() == 2) {
		tunever_device = new Tango::DeviceProxy(resv[0]);
		tunever_attribute = resv[1];
	}

	/* create 2 groups (two threads) to request orbit position to be copied into the simulator */
	vector <string> match_bpm_names[CATSYNC_MATCH_ORB_NUM];
	vector <string> attr_names = {"matchfbx_sensor_names","matchfby_sensor_names"};

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) 
		catdevice->read_attribute(attr_names[i]) >> match_bpm_names[i];
	
	vector <string> grp_names = {"orb_hor","orb_ver"};

	for (int i = 0; i < CATSYNC_MATCH_ORB_NUM; i++) {
		match_orb[i] = new Tango::Group(grp_names[i]);
		for (size_t j = 0; j < match_bpm_names[i].size(); j++) {
			string bpm_name = "sr/diagnostics/" + match_bpm_names[i][j];
			match_orb[i]->add(bpm_name);
			// read position from shared memory
#ifdef DPDK
			if (i == 0) {
				gof2_bpm_idx.push_back(get_bpm_index_from_label(bpm_name.c_str()));
			}
#endif
		}


	}

	// get the list of all magnet/sim_attribute couples
	for (int i = 0; i < magnetConfiguration.size(); i++) {
		
		if ((magnetConfiguration[i][0] == '#') && (magnetConfiguration[i][1] == '#')) 
			break;

		if (magnetConfiguration[i][0] == '#')
			continue;

		vector<string> results;
		string_explode(magnetConfiguration[i], ";", &results);

		INFO_STREAM << __FUNCTION__ << ", " << magnetConfiguration[i] << endl;

		// four parameters magnet_devname;magnet_attrname;sim_attrname;
		std::string magnet_devicename, magnet_attrname, sim_attrname;
		int section_idx, vec_idx;
		double conv_factor;
		
		// magnet_devname;magnet_attrname[X];sim_attrname;section;
		if (results.size() != 5) {
			TangoSys_MemStream error_stream;
			error_stream << "invalid number of MagnetConfiguration parameters: " << magnetConfiguration[i];
			INFO_STREAM << __FUNCTION__ << ": " << error_stream.str() << endl;
			Tango::Except::throw_exception(
				(const char *) "Failed to get MagnetConfiguration",
				error_stream.str(),
				__FUNCTION__, Tango::ERR);
		}

		section_idx = atoi(results[3].c_str());
		if ((section_idx < 0) || (section_idx >= CATSYNC_MAX_THREADS)) {
			TangoSys_MemStream error_stream;
			error_stream << "invalid thread/section number " << results[3];
			INFO_STREAM << __FUNCTION__ << ": " << error_stream.str() << endl;
			Tango::Except::throw_exception(
				(const char *) "Failed to get MagnetConfiguration",
				error_stream.str(),
				__FUNCTION__, Tango::ERR);
		}

		magnet_devicename = results[0];

		try {
			Tango::DeviceProxy *magnetdevice = new Tango::DeviceProxy(magnet_devicename);
		}
		catch (Tango::DevFailed &e) {
			TangoSys_MemStream error_stream;
			error_stream << e.errors[0].desc << ", error connecting to " << magnet_devicename << ", " << e.errors[0].desc;
			INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
			Tango::Except::throw_exception(
				(const char *) "Failed to get MagnetConfiguration",
				error_stream.str(),
				__FUNCTION__, Tango::ERR);
		}
		catch(...) {
			TangoSys_MemStream error_stream;
			error_stream << "Generic error, failed to connect to " << ", " << magnet_devicename;
			INFO_STREAM << __FUNCTION__ << error_stream.str() << endl;
			Tango::Except::throw_exception(
				(const char *) "Failed to get MagnetConfiguration",
				error_stream.str(),
				__FUNCTION__, Tango::ERR);			
		}

		// magnet device; magnet attribute; simulator attribute; thread number
		string attrname(results[1]);		

		size_t open_bracket = attrname.find('[');
		size_t close_bracket = attrname.find(']');

		if (open_bracket != std::string::npos && close_bracket != std::string::npos && close_bracket > open_bracket) {
			// Extract the index as a substring and convert to integer
			std::string index_str = attrname.substr(open_bracket + 1, close_bracket - open_bracket - 1);
			try {
				vec_idx = std::stoi(index_str);  // Convert to integer
			} catch (const std::exception&) {
				vec_idx = 0;  // Default to 0 if conversion fails
				std::cerr << "Invalid index format in source device string. Defaulting to 0." << std::endl;
			}
			// Remove the index from the source device string
			magnet_attrname = attrname.substr(0, open_bracket);
		} 
		else {
			// No index specified, default to 0
			magnet_attrname = attrname;
			vec_idx = 0;
		}

		sim_attrname = results[2];

		conv_factor = std::stod(results[4]);

		// copy data into section (thread) structure
		ps_magnet_sync[section_idx].magnet_devicename.push_back(magnet_devicename);
		// connect to magnet
		Tango::DeviceProxy *device = new Tango::DeviceProxy(magnet_devicename);	
		ps_magnet_sync[section_idx].magnetdevice.push_back(device);
		// get sim attrname
		ps_magnet_sync[section_idx].magnet_attrname.push_back(magnet_attrname);		
		ps_magnet_sync[section_idx].sim_attrname.push_back(sim_attrname);	
		ps_magnet_sync[section_idx].vec_idx.push_back(vec_idx);
		ps_magnet_sync[section_idx].conv_factor.push_back(conv_factor);

		// check if the start of the sync of the ps has to be delayed
		struct timespec delay;
		delay.tv_sec = 0;
		delay.tv_nsec = 0;
		for (int j = 0; j < psDelayedStartList.size(); j++) {
			//cout << psDelayedStartList[j] << " " << magnet_devicename << endl;
			vector<string> res;
			string_explode(psDelayedStartList[j], ";", &res);
			if (res.size() == 2) {
				if (match(magnet_devicename.c_str(), res[0].c_str()) == true)  {
					int delay_int = atoi(res[1].c_str());
					if (delay_int > delay.tv_sec) {
						delay.tv_sec = delay_int;
					}
				}
			}
		}
		ps_magnet_sync[section_idx].start_time.push_back(delay); // initialize with unuseful value
		ps_magnet_sync[section_idx].delay_start.push_back(delay);

		INFO_STREAM << __FUNCTION__ << ", magnet devicename: " << magnet_devicename << endl;
		INFO_STREAM << __FUNCTION__ << ", magnet attribute: " << magnet_attrname << endl;
		INFO_STREAM << __FUNCTION__ << ", magnet attribute index: " << vec_idx << endl;
		INFO_STREAM << __FUNCTION__ << ", magnet conversion factor: " << conv_factor << endl;
		INFO_STREAM << __FUNCTION__ << ", magnet section thread index: " << section_idx << endl;
		INFO_STREAM << __FUNCTION__ << ", simulator attribute: " << sim_attrname << endl;
		INFO_STREAM << __FUNCTION__ << ", start delay (s): " << delay.tv_sec << endl;

	}

	// pre feedback sequencers configuration
	for (size_t i = 0; i < preSequencer.size(); i++) {
		if (preSequencer[i].size() > 0) {
			INFO_STREAM << "Creating pre feedback sequence: " << preSequencer[i];
			Sequencer sequencer_obj(preSequencer[i]);
			pre_sequencer_v.push_back(sequencer_obj);
			INFO_STREAM << "OK" << endl;
		}
	}
	// post feedback sequencers configuration
	for (size_t i = 0; i < postSequencer.size(); i++) {
		if (postSequencer[i].size() > 0) {
			INFO_STREAM << "Creating post feedback sequence: " << postSequencer[i];
			Sequencer sequencer_obj(postSequencer[i]);
			post_sequencer_v.push_back(sequencer_obj);
			INFO_STREAM << "OK" << endl;
		}
	}
	
	for (int i = 0; i < mIMOFBList.size(); i++) {
		Tango::DeviceProxy *fbdevice = new Tango::DeviceProxy(mIMOFBList[i]);
		if (fbdevice != NULL)
			mimofbdevice.push_back(fbdevice);
	}


}


//--------------------------------------------------------
/**
 *	Method      : CATSync::cleanup_devices()
 * Description:  Cleanup tango devices and data structures
 *                
 */
//--------------------------------------------------------
void CATSync::cleanup_devices()
{

	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		// deallocate all device proxies		
		for (Tango::DeviceProxy* obj :  ps_magnet_sync[i].magnetdevice) {
			if (obj)
				delete obj;
		}
		ps_magnet_sync[i].magnetdevice.clear();
		ps_magnet_sync[i].magnet_devicename.clear();		
		ps_magnet_sync[i].magnet_attrname.clear();
		ps_magnet_sync[i].sim_attrname.clear();
		ps_magnet_sync[i].vec_idx.clear();
		ps_magnet_sync[i].conv_factor.clear();
	}

	if (catdevice)
		delete catdevice;

	if (hlsdevice)
		delete hlsdevice;

	if (energydevice)
		delete energydevice;

	if (tunehor_device)
		delete tunehor_device;

	if (tunever_device)
		delete tunever_device;

	pre_sequencer_v.clear();
	post_sequencer_v.clear();

	min_energies.clear();
	max_energies.clear();

	source_point_attributes.clear();

	for (auto ptr : mimofbdevice) {
		delete ptr; // free pointer
	}
	mimofbdevice.clear(); // clear vector


}

void CATSync::start_sequencer(int type) 
{
	if (run_sequencer_flag[type] == false) {
		abort_sequencer_flag[type] = false;
		err_sequencer_flag[type] = false;	
		sequencerloop[type] = new sequencerthread(this,type);		
		sequencerloop[type]->start();
	}
}


int32_t CATSync::get_bpm_index_from_label(const char *input_string)
{
	int bpm_indexes[2];
	int num_indexes;

	extract_numbers(input_string, bpm_indexes, &num_indexes, 2);
#ifdef DPDK	
	for (int32_t i = 0; i < GOF2_NUM_BPM_ACTIVED; i++) {
		if ((bpm_indexes[0] == bpm_addr_tb[i][0]) && (bpm_indexes[1] == bpm_addr_tb[i][1]))
			return bpm_addr_tb[i][4];
	}

	Tango::Except::throw_exception(
		(const char *)"Failed to get bpm index",
		(const char *)"Bpm name not found",
		__FUNCTION__);
#endif
}


void CATSync::extract_numbers(const char *str, int numbers[], int *count, int max_count) 
{
	*count = 0; 
	const char *ptr = str;

	while (*ptr) {
		if (isdigit(*ptr)) {
			numbers[*count] = strtol(ptr, (char **)&ptr, 10);
			(*count)++; 
			if ((*count) == max_count) // expect only two numbers in the bpm label; ex: bpm_s1.2
				break;
		} else {
			ptr++; 
		}
	}
}


void CATSync::device_set_energy(double energy, int num_thread) 
{

	for (int i = 0; i < ps_magnet_sync[num_thread].magnetdevice.size(); i++) {
		try { 
			double cur_energy;
			ps_magnet_sync[num_thread].magnetdevice[i]->read_attribute("Energy") >> cur_energy;
			if (cur_energy != energy) {
				Tango::DeviceAttribute attr("Energy", energy);
				ps_magnet_sync[num_thread].magnetdevice[i]->write_attribute(attr);
			}
		}
		catch (Tango::DevFailed &e) {
			TangoSys_MemStream error_stream;
			error_stream << e.errors[0].desc << ", error setting " << ps_magnet_sync[num_thread].magnet_devicename[i] << ", " << e.errors[0].desc;
			INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
		}
		catch(...){}
	}

	if ((num_thread == CATSYNC_MAX_THREADS - 1) && (*attr_EnableCATEnergyUpdate_read)) {
		try {
			double cur_energy;
			catdevice->read_attribute("Energy") >> cur_energy;
			if (cur_energy != energy) {
				Tango::DeviceAttribute attr("Energy", energy*1000000000); // simulator works with eV
				catdevice->write_attribute(attr);
			}
		}
		catch (Tango::DevFailed &e) {
			TangoSys_MemStream error_stream;
			error_stream << e.errors[0].desc << ", error setting " << cATDSDevice << ", " << e.errors[0].desc;
			INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
		}
		catch(...) {}	
	}

}


void CATSync::device_update_energy(int num_thread) 
{
	double energy;
	energydevice->read_attribute(energy_attribute.c_str()) >> energy;

	// to test energy change
	//double random_value = (double)rand() / (double)RAND_MAX;
	//energy += 0.123+random_value*0.01;

	if (energy != old_energy[num_thread]) {
		for (int i = 0; i < min_energies.size(); i++) {
			if ((energy >= min_energies[i]) && (energy <= max_energies[i])) {
				device_set_energy(energy, num_thread);
				break;
			}
		}
	}

	old_energy[num_thread] = energy;

}


void CATSync::synchronize_attributes(void)
{
	for (const auto &attr : source_point_attributes) {
		try {
			// first convert sourcepoints m/mrad to mm/mmrad
			string attr_g = attr + "_errgain";
			Tango::DeviceAttribute dev_attr_g(attr_g, (double)1000);
			catdevice->write_attribute(dev_attr_g);

			// Read attribute from the old device
			vector <double> hls_value, cat_value, err_offset;
			// get sourcepoint values from hls and cat
			string attr_sp = attr;
			hlsdevice->read_attribute(attr_sp) >> hls_value;
			catdevice->read_attribute(attr_sp) >> cat_value;

			string attr_r = attr + "_erroffset";
			catdevice->read_attribute(attr_r) >> err_offset;

			// Calculate the error offset
			double new_err_offset = (hls_value[0] - cat_value[0]) + err_offset[0];

			// Write error offset and set gain to 1000 on the new device
			string attr_w = attr + "_erroffset";
			Tango::DeviceAttribute dev_attr_w(attr_w, new_err_offset);

			catdevice->write_attribute(dev_attr_w);

			INFO_STREAM << "Synchronized " << attr << ", hls=" << hls_value[0] << ", cat=" << cat_value[0]
				<< ", offset_old=" << err_offset[0] << ", offset_new=" << new_err_offset << endl;
		} 
		catch (Tango::DevFailed &e) {
			INFO_STREAM << "Error processing attribute " << attr << ": " << e.errors[0].desc << std::endl;
		}
	}
}


void CATSync::convert_sourcepoints_mm(void)
{
	for (const auto &attr : source_point_attributes) {
		try {
			// first convert sourcepoints m/mrad to mm/mmrad
			string attr_g = attr + "_errgain";
			Tango::DeviceAttribute dev_attr_g(attr_g, (double)1000);
			catdevice->write_attribute(dev_attr_g);
		} 
		catch (Tango::DevFailed &e) {
			INFO_STREAM << "Error processing attribute " << attr << ": " << e.errors[0].desc << std::endl;
		}
	}
}


void CATSync::reset_sourcepoint_offsets(void)
{
	for (const auto &attr : source_point_attributes) {
		try {
			string attr_g = attr + "_erroffset";
			Tango::DeviceAttribute dev_attr_g(attr_g, (double)0);
			catdevice->write_attribute(dev_attr_g);
		} 
		catch (Tango::DevFailed &e) {
			INFO_STREAM << "Error processing attribute " << attr << ": " << e.errors[0].desc << std::endl;
		}
	}
}


void CATSync::switchString(std::string &source, const std::string &from, const std::string &to, bool reverse) 
{
	if (reverse) {
		size_t start_pos = 0;
		while ((start_pos = source.find(to, start_pos)) != std::string::npos) {
			source.replace(start_pos, to.length(), from);
			start_pos += from.length();
		}
	} 
	else {
		size_t start_pos = 0;
		while ((start_pos = source.find(from, start_pos)) != std::string::npos) {
			source.replace(start_pos, from.length(), to);
			start_pos += to.length();
		}
	}
}


void CATSync::updateSensorConfiguration(bool reverse) 
{
	try {
		for (const auto device : mimofbdevice) {
 			// Prepare the SensorConfiguration property to read
			Tango::DbData db_data;
			db_data.push_back(Tango::DbDatum("SensorConfiguration"));

			// Read the SensorConfiguration property
			device->get_property(db_data);

			// Extract the property value (array of strings)
			std::vector<std::string> sensorConfig;
			db_data[0] >> sensorConfig;

			// Update the strings in the property
			for (auto &config : sensorConfig) {
				switchString(config, "/bpmid;", "/bpmid_new;", reverse);
			}

			// Write the updated property back to the Tango database
			Tango::DbDatum updatedProperty("SensorConfiguration");
			updatedProperty << sensorConfig;
			Tango::DbData db_data_put;
			db_data_put.push_back(updatedProperty);
			device->put_property(db_data_put);

			device->command_inout("Init");

			// set time factor
			double time_factor;
			if (reverse == false)	
				time_factor = *attr_LFBCATTimeFactor_read;
			else
				time_factor = *attr_LFBHLSTimeFactor_read;

			Tango::DeviceAttribute attr("TimeFactor", time_factor);
			device->write_attribute(attr);

		}
	} 
	catch (const Tango::DevFailed &e) {
		INFO_STREAM << __FUNCTION__ << ", tango exception: " << e.errors[0].desc << std::endl;
	} 
	catch (const std::exception &e) {
		INFO_STREAM << __FUNCTION__ << ", standard exception: " << e.what() << std::endl;
	}
}


bool CATSync::match(const char* str, const char* pattern)
{

    if ((str == NULL) || (pattern == NULL))
        return false;

    if (*pattern == '\0') {
        return (*str == '\0');
    }

    if (*pattern == '*') {
        while (*str != '\0') {
            if (match(str, pattern+1)) {
                return true;
            }
            str++;
        }
        return match(str, pattern+1);
    }

    if (*str == '\0') {
        return false;
    }

    if (*pattern == *str) {
        return match(str+1, pattern+1);
    }

    return false;
}



/* clang-format off */// //--------------------------------------------------------
// /**
//  *	Read attribute ThreadStates related method
//  *
//  *
//  *	Data type:	Tango::DevState
//  *	Attr type:	Spectrum max = 10000
//  */
// //--------------------------------------------------------
// void CATSync::read_ThreadStates(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_ThreadStates(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_ThreadStates_read, CATSYNC_MAX_THREADS+CATSYNC_MATCH_ORB_NUM);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute ThreadSettings related method
//  *
//  *
//  *	Data type:	Tango::DevULong
//  *	Attr type:	Spectrum max = 10000
//  */
// //--------------------------------------------------------
// void CATSync::read_ThreadSettings(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_ThreadSettings(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_ThreadSettings_read, CATSYNC_MAX_THREADS+CATSYNC_MATCH_ORB_NUM);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute SyncThreadErrors related method
//  *
//  *
//  *	Data type:	Tango::DevULong
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::read_SyncThreadErrors(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_SyncThreadErrors(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_SyncThreadErrors_read, SyncThreadErrorsAttrib::X_DATA_SIZE);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute EnableCATMatchOrbit related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::read_EnableCATMatchOrbit(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_EnableCATMatchOrbit(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_EnableCATMatchOrbit_read);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Write attribute EnableCATMatchOrbit related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::write_EnableCATMatchOrbit(Tango::WAttribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::write_EnableCATMatchOrbit(Tango::WAttribute &attr) entering... " << std::endl;
// 	//	Retrieve write value
// 	Tango::DevBoolean	w_val;
// 	attr.get_write_value(w_val);
// 	/* clang-format on */
// 	//	Add your own code
// 	*attr_EnableCATMatchOrbit_read = w_val;
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute EnableCATMatchOrbit related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::read_EnableCATMatchOrbit(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_EnableCATMatchOrbit(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_EnableCATMatchOrbit_read);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Write attribute EnableCATMatchOrbit related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::write_EnableCATMatchOrbit(Tango::WAttribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::write_EnableCATMatchOrbit(Tango::WAttribute &attr) entering... " << std::endl;
// 	//	Retrieve write value
// 	Tango::DevBoolean	w_val;
// 	attr.get_write_value(w_val);
// 	/* clang-format on */
// 	//	Add your own code
// 	*attr_EnableCATMatchOrbit_read = w_val;
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute EnableCATSyncMagnets related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::read_EnableCATSyncMagnets(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_EnableCATSyncMagnets(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_EnableCATSyncMagnets_read);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Write attribute EnableCATSyncMagnets related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::write_EnableCATSyncMagnets(Tango::WAttribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::write_EnableCATSyncMagnets(Tango::WAttribute &attr) entering... " << std::endl;
// 	//	Retrieve write value
// 	Tango::DevBoolean	w_val;
// 	attr.get_write_value(w_val);
// 	/* clang-format on */
// 	//	Add your own code
// 	*attr_EnableCATSyncMagnets_read = w_val;
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute EnableCATSourcePointScaling related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::read_EnableCATSourcePointScaling(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::read_EnableCATSourcePointScaling(Tango::Attribute &attr) entering... " << std::endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_EnableCATSourcePointScaling_read);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Write attribute EnableCATSourcePointScaling related method
//  *
//  *
//  *	Data type:	Tango::DevBoolean
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void CATSync::write_EnableCATSourcePointScaling(Tango::WAttribute &attr)
// {
// 	DEBUG_STREAM << "CATSync::write_EnableCATSourcePointScaling(Tango::WAttribute &attr) entering... " << std::endl;
// 	//	Retrieve write value
// 	Tango::DevBoolean	w_val;
// 	attr.get_write_value(w_val);
// 	/* clang-format on */
// 	//	Add your own code
// 	/* clang-format off */
// }


/*----- PROTECTED REGION END -----*/	//	CATSync::namespace_ending
} //	namespace
