/*----- PROTECTED REGION ID(CATSync.h) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        CATSync.h
//
// description : Include file for the CATSync class
//
// project :     CATSync
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef CATSync_H
#define CATSync_H

#include <tango/tango.h>
#include "Sequencer.h"
#include "util.h"

#ifdef DPDK
	#include <gof2_shm.h>
	#include <gof2_fun.h>
#else
	#define	GOF2_BUF_SIZE 2000000
#endif

#define CATSYNC_MAX_THREADS 13 // 12 sections + bigps
#define CATSYNC_MATCH_ORBHOR_IDX 0
#define CATSYNC_MATCH_ORBVER_IDX 1
#define CATSYNC_MATCH_ORB_NUM 2

#define CATSYNC_PRESEQUENCER_IDX 0
#define CATSYNC_POSTSEQUENCER_IDX 1
#define CATSYNC_SEQUENCER_NUM 2

#define CATSYNC_MAX_PS_SECTION 30 // 30 PS
#define CATSYNC_LEN 1024

#define CATSYNC_ENERGY_TOUT 10 // seconds

class Sequencer;

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync.h

#ifdef TANGO_LOG
	// cppTango after c934adea (Merge branch 'remove-cout-definition' into 'main', 2022-05-23)
	// nothing to do
#else
	// cppTango 9.3-backports and older
	#define TANGO_LOG       cout
	#define TANGO_LOG_INFO  cout2
	#define TANGO_LOG_DEBUG cout3
#endif // TANGO_LOG

/**
 *  CATSync class description:
 *    Tango server used to synchronize CAT simulator with field
 */


namespace CATSync_ns
{
/*----- PROTECTED REGION ID(CATSync::Additional Class Declarations) ENABLED START -----*/
/* clang-format on */
//	Additional Class Declarations

	class matchthread;
	class catsyncthread;
	class sequencerthread;

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync::Additional Class Declarations

class CATSync : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(CATSync::Data Members) ENABLED START -----*/
/* clang-format on */

	friend class matchthread;
	friend class catsyncthread;
	friend class sequencerthread;
	friend class Sequencer;

//	Add your own data members

public:


	int fd_shm;
	void *shm_addr;
	struct gof2_shm_t *gof2;
	bool enable_hugepages;


	struct ps_magnet_sync_t  
	{
		vector <string> sim_attrname;
		vector <Tango::DeviceProxy *> magnetdevice;
		vector <string> magnet_devicename;
		vector <string> magnet_attrname;
		vector <int> section_idx;	
		vector <double> conv_factor;
		vector <double> value;
		vector <double> old_value;
		vector <int> vec_idx;	
		vector <struct timespec> delay_start;
		vector <struct timespec> start_time;
	};

	Tango::Group *match_orb[CATSYNC_MATCH_ORB_NUM]; 
	
	struct timespec orbit_start_time[CATSYNC_MATCH_ORB_NUM];
	struct timespec loop_start_time;
	struct timespec init_cat_time;

	catsyncthread *catsyncloop[CATSYNC_MAX_THREADS];
	matchthread *matchloop[CATSYNC_MATCH_ORB_NUM];
	sequencerthread *sequencerloop[CATSYNC_SEQUENCER_NUM];

	bool abort_catsync_flag[CATSYNC_MAX_THREADS], 
	     run_catsync_flag[CATSYNC_MAX_THREADS],
             err_catsync_flag[CATSYNC_MAX_THREADS];

	bool abort_sequencer_flag[CATSYNC_SEQUENCER_NUM], 
	     run_sequencer_flag[CATSYNC_SEQUENCER_NUM],
             err_sequencer_flag[CATSYNC_SEQUENCER_NUM];

	bool abort_match_orb_flag[CATSYNC_MATCH_ORB_NUM], 
	     run_match_orb_flag[CATSYNC_MATCH_ORB_NUM],
	     err_match_orb_flag[CATSYNC_MATCH_ORB_NUM];

	struct ps_magnet_sync_t ps_magnet_sync[CATSYNC_MAX_THREADS];

	Tango::DeviceProxy *catdevice;
	Tango::DeviceProxy *hlsdevice;
	vector <Tango::DeviceProxy *> mimofbdevice;

	Tango::DeviceProxy *energydevice;
	string energy_attribute;
	vector <double> min_energies, max_energies;
	vector <string> source_point_attributes;
	double old_energy[CATSYNC_MAX_THREADS];
	struct timespec energy_update_time[CATSYNC_MAX_THREADS], source_points_update_time;

	vector <Sequencer> pre_sequencer_v;
	vector <Sequencer> post_sequencer_v;

	Tango::DeviceProxy *tunehor_device;
	Tango::DeviceProxy *tunever_device;
	string tunehor_attribute, tunever_attribute;

	vector <int32_t> gof2_bpm_idx;

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync::Data Members

//	Device property data members
public:
	//	MagnetConfiguration:	device_read_attr[];device_set_attr;thread_num(0,1,...N)
	//  device_read_attr[];device_set_attr;thread_num(0,1,...N)
	std::vector<std::string>	magnetConfiguration;
	//	PreSequencer:	
	std::vector<std::string>	preSequencer;
	//	PostSequencer:	
	std::vector<std::string>	postSequencer;
	//	AutoStart:	
	Tango::DevBoolean	autoStart;
	//	CATDSDevice:	
	std::string	cATDSDevice;
	//	HugePages:	
	Tango::DevBoolean	hugePages;
	//	SharedMemoryID:	
	std::string	sharedMemoryID;
	//	EnergyDevice:	Device;attribute
	std::string	energyDevice;
	//	EnergyRanges:	Range in which the energy is at every change is written to magnets and simulators.
	//  MinEnergy1;MaxEnergy1
	//  MinEnergy2;MaxEnergy2
	std::vector<std::string>	energyRanges;
	//	HLSDevice:	
	std::string	hLSDevice;
	//	SourcePointLabels:	X_S1,Y_S1,XP_S1,YP_S1
	//  X_S2,Y_S2,XP_S2,YP_S2
	//  ......
	std::vector<std::string>	sourcePointLabels;
	//	EnergyLoopPeriodSec:	Period in sec. of checking energy change and updating simulator and magnet energies
	Tango::DevLong	energyLoopPeriodSec;
	//	SyncLoopPeriodSec:	Period in sec. of synching CAT simulato rwith HLS in source points by modifying CAT source point gain and offset
	Tango::DevLong	syncLoopPeriodSec;
	//	MIMOFBList:	
	std::vector<std::string>	mIMOFBList;
	//	PsDelayedStartList:	matched string; delay in seconds
	//  Ex: *Q2_s*;3
	std::vector<std::string>	psDelayedStartList;
	//	OrbitDelayedStartList:	delay in seconds for horizontal plane
	//  delay in second for vertical plane
	std::vector<Tango::DevLong>	orbitDelayedStartList;
	//	TuneHorDevice:	tunedevice;tune_attribute
	std::string	tuneHorDevice;
	//	TuneVerDevice:	tunedevice;tune_attribute
	std::string	tuneVerDevice;

//	Attribute data members
public:
	Tango::DevBoolean	*attr_EnableCATEnergyUpdate_read;
	Tango::DevBoolean	*attr_EnableCATSourcePointUpdate_read;
	Tango::DevDouble	*attr_LFBCATTimeFactor_read;
	Tango::DevDouble	*attr_LFBHLSTimeFactor_read;
	Tango::DevBoolean	*attr_EnableCATMatchOrbit_read;
	Tango::DevBoolean	*attr_EnableCATSyncMagnets_read;
	Tango::DevState	*attr_CATState_read;
	Tango::DevState	*attr_HLSState_read;
	Tango::DevDouble	*attr_SyncGain_read;
	Tango::DevDouble	*attr_CATTuneHor_read;
	Tango::DevDouble	*attr_CATTuneVer_read;
	Tango::DevString	*attr_MatchThreadErrMsg_read;
	Tango::DevString	*attr_SyncThreadErrMsg_read;
	Tango::DevULong	*attr_MatchThreadErrCnt_read;
	Tango::DevULong	*attr_SyncThreadErrCnt_read;
	Tango::DevULong	*attr_MatchThreadCnt_read;
	Tango::DevULong	*attr_SyncThreadCnt_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	CATSync(Tango::DeviceClass *cl,std::string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	CATSync(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	CATSync(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */
	~CATSync();


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();


//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method      : CATSync::read_attr_hardware()
	 * Description:  Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(std::vector<long> &attr_list);
	//--------------------------------------------------------
	/*
	 *	Method      : CATSync::write_attr_hardware()
	 * Description:  Hardware writing for attributes.
	 */
	//--------------------------------------------------------
	virtual void write_attr_hardware(std::vector<long> &attr_list);

/**
 *	Attribute EnableCATEnergyUpdate related methods
 *
 *
 *	Data type:  Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_EnableCATEnergyUpdate(Tango::Attribute &attr);
	virtual void write_EnableCATEnergyUpdate(Tango::WAttribute &attr);
	virtual bool is_EnableCATEnergyUpdate_allowed(Tango::AttReqType type);
/**
 *	Attribute EnableCATSourcePointUpdate related methods
 *
 *
 *	Data type:  Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_EnableCATSourcePointUpdate(Tango::Attribute &attr);
	virtual void write_EnableCATSourcePointUpdate(Tango::WAttribute &attr);
	virtual bool is_EnableCATSourcePointUpdate_allowed(Tango::AttReqType type);
/**
 *	Attribute LFBCATTimeFactor related methods
 *
 *
 *	Data type:  Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_LFBCATTimeFactor(Tango::Attribute &attr);
	virtual void write_LFBCATTimeFactor(Tango::WAttribute &attr);
	virtual bool is_LFBCATTimeFactor_allowed(Tango::AttReqType type);
/**
 *	Attribute LFBHLSTimeFactor related methods
 *
 *
 *	Data type:  Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_LFBHLSTimeFactor(Tango::Attribute &attr);
	virtual void write_LFBHLSTimeFactor(Tango::WAttribute &attr);
	virtual bool is_LFBHLSTimeFactor_allowed(Tango::AttReqType type);
/**
 *	Attribute EnableCATMatchOrbit related methods
 *
 *
 *	Data type:  Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_EnableCATMatchOrbit(Tango::Attribute &attr);
	virtual void write_EnableCATMatchOrbit(Tango::WAttribute &attr);
	virtual bool is_EnableCATMatchOrbit_allowed(Tango::AttReqType type);
/**
 *	Attribute EnableCATSyncMagnets related methods
 *
 *
 *	Data type:  Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_EnableCATSyncMagnets(Tango::Attribute &attr);
	virtual void write_EnableCATSyncMagnets(Tango::WAttribute &attr);
	virtual bool is_EnableCATSyncMagnets_allowed(Tango::AttReqType type);
/**
 *	Attribute CATState related methods
 *
 *
 *	Data type:  Tango::DevState
 *	Attr type:	Scalar
 */
	virtual void read_CATState(Tango::Attribute &attr);
	virtual bool is_CATState_allowed(Tango::AttReqType type);
/**
 *	Attribute HLSState related methods
 *
 *
 *	Data type:  Tango::DevState
 *	Attr type:	Scalar
 */
	virtual void read_HLSState(Tango::Attribute &attr);
	virtual bool is_HLSState_allowed(Tango::AttReqType type);
/**
 *	Attribute SyncGain related methods
 *
 *
 *	Data type:  Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_SyncGain(Tango::Attribute &attr);
	virtual void write_SyncGain(Tango::WAttribute &attr);
	virtual bool is_SyncGain_allowed(Tango::AttReqType type);
/**
 *	Attribute CATTuneHor related methods
 *
 *
 *	Data type:  Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CATTuneHor(Tango::Attribute &attr);
	virtual bool is_CATTuneHor_allowed(Tango::AttReqType type);
/**
 *	Attribute CATTuneVer related methods
 *
 *
 *	Data type:  Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_CATTuneVer(Tango::Attribute &attr);
	virtual bool is_CATTuneVer_allowed(Tango::AttReqType type);
/**
 *	Attribute MatchThreadErrMsg related methods
 *
 *
 *	Data type:  Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_MatchThreadErrMsg(Tango::Attribute &attr);
	virtual bool is_MatchThreadErrMsg_allowed(Tango::AttReqType type);
/**
 *	Attribute SyncThreadErrMsg related methods
 *
 *
 *	Data type:  Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_SyncThreadErrMsg(Tango::Attribute &attr);
	virtual bool is_SyncThreadErrMsg_allowed(Tango::AttReqType type);
/**
 *	Attribute MatchThreadErrCnt related methods
 *
 *
 *	Data type:  Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_MatchThreadErrCnt(Tango::Attribute &attr);
	virtual bool is_MatchThreadErrCnt_allowed(Tango::AttReqType type);
/**
 *	Attribute SyncThreadErrCnt related methods
 *
 *
 *	Data type:  Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_SyncThreadErrCnt(Tango::Attribute &attr);
	virtual bool is_SyncThreadErrCnt_allowed(Tango::AttReqType type);
/**
 *	Attribute MatchThreadCnt related methods
 *
 *
 *	Data type:  Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_MatchThreadCnt(Tango::Attribute &attr);
	virtual bool is_MatchThreadCnt_allowed(Tango::AttReqType type);
/**
 *	Attribute SyncThreadCnt related methods
 *
 *
 *	Data type:  Tango::DevULong
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_SyncThreadCnt(Tango::Attribute &attr);
	virtual bool is_SyncThreadCnt_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method      : CATSync::add_dynamic_attributes()
	 * Description:  Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();




//	Command related methods
public:
	/**
	 *	Command Start related method
	 *
	 *
	 */
	virtual void start();
	virtual bool is_Start_allowed(const CORBA::Any &any);
	/**
	 *	Command Stop related method
	 *
	 *
	 */
	virtual void stop();
	virtual bool is_Stop_allowed(const CORBA::Any &any);
	/**
	 *	Command GotoEnergy related method
	 *
	 *
	 *	@param argin Energy (GeV)
	 */
	virtual void goto_energy(Tango::DevDouble argin);
	virtual bool is_GotoEnergy_allowed(const CORBA::Any &any);
	/**
	 *	Command UpdateEnergy related method
	 *
	 *
	 */
	virtual void update_energy();
	virtual bool is_UpdateEnergy_allowed(const CORBA::Any &any);
	/**
	 *	Command SyncSourcePointsToHLS related method
	 *
	 *
	 */
	virtual void sync_source_points_to_hls();
	virtual bool is_SyncSourcePointsToHLS_allowed(const CORBA::Any &any);
	/**
	 *	Command SwitchLFBCAT related method
	 *
	 *
	 */
	virtual void switch_lfbcat();
	virtual bool is_SwitchLFBCAT_allowed(const CORBA::Any &any);
	/**
	 *	Command SwitchLFBHLS related method
	 *
	 *
	 */
	virtual void switch_lfbhls();
	virtual bool is_SwitchLFBHLS_allowed(const CORBA::Any &any);
	/**
	 *	Command ConvertSourcePointsMM related method
	 *
	 *
	 */
	virtual void convert_source_points_mm();
	virtual bool is_ConvertSourcePointsMM_allowed(const CORBA::Any &any);
	/**
	 *	Command ResetSourcePointOffsets related method
	 *
	 *
	 */
	virtual void reset_source_point_offsets();
	virtual bool is_ResetSourcePointOffsets_allowed(const CORBA::Any &any);


	//--------------------------------------------------------
	/**
	 *	Method      : CATSync::add_dynamic_commands()
	 * Description:  Add dynamic commands if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_commands();

/*----- PROTECTED REGION ID(CATSync::Additional Method prototypes) ENABLED START -----*/
/* clang-format on */
//	Additional Method prototypes

void connect_devices();
void cleanup_devices();
void start_sequencer(int type);
void extract_numbers(const char *str, int numbers[], int *count, int max_count);
int32_t get_bpm_index_from_label(const char *input_string);
void device_set_energy(double energy, int num_thread);
void device_update_energy(int num_thread);
void synchronize_attributes(void);
void switchString(std::string &source, const std::string &from, const std::string &to, bool reverse);
void updateSensorConfiguration(bool reverse);
bool match(const char* str, const char* pattern);
void convert_sourcepoints_mm(void);
void reset_sourcepoint_offsets(void);

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync::Additional Method prototypes
};

/*----- PROTECTED REGION ID(CATSync::Additional Classes Definitions) ENABLED START -----*/
/* clang-format on */
//	Additional Classes Definitions
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSync::Additional Classes Definitions

}	//	End of namespace

#endif   //	CATSync_H
