/*----- PROTECTED REGION ID(CATSyncClass.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        CATSyncClass.cpp
//
// description : C++ source for the CATSyncClass.
//               A singleton class derived from DeviceClass.
//               It implements the command and attribute list
//               and all properties and methods required
//               by the CATSync once per process.
//
// project :     CATSync
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "CATSyncClass.h"
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSyncClass.cpp

//-------------------------------------------------------------------
/**
 *	Create CATSyncClass singleton and
 *	return it in a C function for Python usage
 */
//-------------------------------------------------------------------
extern "C" {
#ifdef _TG_WINDOWS_

__declspec(dllexport)

#endif

	Tango::DeviceClass *_create_CATSync_class(const char *name) {
		return CATSync_ns::CATSyncClass::init(name);
	}
}

namespace CATSync_ns
{
//===================================================================
//	Initialize pointer for singleton pattern
//===================================================================
CATSyncClass *CATSyncClass::_instance = NULL;

//===================================================================
//	Class constants
//===================================================================
constexpr long MatchThreadErrMsgAttrib::X_DATA_SIZE;
constexpr long SyncThreadErrMsgAttrib::X_DATA_SIZE;
constexpr long MatchThreadErrCntAttrib::X_DATA_SIZE;
constexpr long SyncThreadErrCntAttrib::X_DATA_SIZE;
constexpr long MatchThreadCntAttrib::X_DATA_SIZE;
constexpr long SyncThreadCntAttrib::X_DATA_SIZE;
//--------------------------------------------------------
/**
 * method : 		CATSyncClass::CATSyncClass(std::string &s)
 * description : 	constructor for the CATSyncClass
 *
 * @param s	The class name
 */
//--------------------------------------------------------
CATSyncClass::CATSyncClass(std::string &s):Tango::DeviceClass(s)
{
	TANGO_LOG_INFO << "Entering CATSyncClass constructor" << std::endl;
	set_default_property();
	write_class_property();

	/*----- PROTECTED REGION ID(CATSyncClass::constructor) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::constructor

	TANGO_LOG_INFO << "Leaving CATSyncClass constructor" << std::endl;
}

//--------------------------------------------------------
/**
 * method : 		CATSyncClass::~CATSyncClass()
 * description : 	destructor for the CATSyncClass
 */
//--------------------------------------------------------
CATSyncClass::~CATSyncClass()
{
	/*----- PROTECTED REGION ID(CATSyncClass::destructor) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::destructor

	_instance = NULL;
}


//--------------------------------------------------------
/**
 * method : 		CATSyncClass::init
 * description : 	Create the object if not already done.
 *                  Otherwise, just return a pointer to the object
 *
 * @param	name	The class name
 */
//--------------------------------------------------------
CATSyncClass *CATSyncClass::init(const char *name)
{
	if (_instance == NULL)
	{
		try
		{
			std::string s(name);
			_instance = new CATSyncClass(s);
		}
		catch (std::bad_alloc &)
		{
			throw;
		}
	}
	return _instance;
}

//--------------------------------------------------------
/**
 * method : 		CATSyncClass::instance
 * description : 	Check if object already created,
 *                  and return a pointer to the object
 */
//--------------------------------------------------------
CATSyncClass *CATSyncClass::instance()
{
	if (_instance == NULL)
	{
		std::cerr << "Class is not initialised !!" << std::endl;
		exit(-1);
	}
	return _instance;
}



//===================================================================
//	Command execution method calls
//===================================================================
//--------------------------------------------------------
/**
 * method : 		StartClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *StartClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "StartClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->start());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		StopClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *StopClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "StopClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->stop());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		GotoEnergyClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GotoEnergyClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GotoEnergyClass::execute(): arrived" << std::endl;
	Tango::DevDouble argin;
	extract(in_any, argin);
	((static_cast<CATSync *>(device))->goto_energy(argin));
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		UpdateEnergyClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *UpdateEnergyClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "UpdateEnergyClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->update_energy());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		SyncSourcePointsToHLSClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *SyncSourcePointsToHLSClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "SyncSourcePointsToHLSClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->sync_source_points_to_hls());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		SwitchLFBCATClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *SwitchLFBCATClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "SwitchLFBCATClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->switch_lfbcat());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		SwitchLFBHLSClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *SwitchLFBHLSClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "SwitchLFBHLSClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->switch_lfbhls());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		ConvertSourcePointsMMClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *ConvertSourcePointsMMClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "ConvertSourcePointsMMClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->convert_source_points_mm());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		ResetSourcePointOffsetsClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *ResetSourcePointOffsetsClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "ResetSourcePointOffsetsClass::execute(): arrived" << std::endl;
	((static_cast<CATSync *>(device))->reset_source_point_offsets());
	return new CORBA::Any();
}


//===================================================================
//	Properties management
//===================================================================
//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::get_class_property()
 * Description:  Get the class property for specified name.
 */
//--------------------------------------------------------
Tango::DbDatum CATSyncClass::get_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_prop.size() ; i++)
		if (cl_prop[i].name == prop_name)
			return cl_prop[i];
	//	if not found, returns  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::get_default_device_property()
 * Description:  Return the default value for device property.
 */
//--------------------------------------------------------
Tango::DbDatum CATSyncClass::get_default_device_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<dev_def_prop.size() ; i++)
		if (dev_def_prop[i].name == prop_name)
			return dev_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::get_default_class_property()
 * Description:  Return the default value for class property.
 */
//--------------------------------------------------------
Tango::DbDatum CATSyncClass::get_default_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_def_prop.size() ; i++)
		if (cl_def_prop[i].name == prop_name)
			return cl_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}


//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::set_default_property()
 * Description:  Set default property (class and device) for wizard.
 *                For each property, add to wizard property name and description.
 *                If default value has been set, add it to wizard property and
 *                store it in a DbDatum.
 */
//--------------------------------------------------------
void CATSyncClass::set_default_property()
{
	std::string	prop_name;
	std::string	prop_desc;
	std::string	prop_def;
	std::vector<std::string>	vect_data;

	//	Set Default Class Properties

	//	Set Default device Properties
	prop_name = "MagnetConfiguration";
	prop_desc = "device_read_attr[];device_set_attr;thread_num(0,1,...N)\ndevice_read_attr[];device_set_attr;thread_num(0,1,...N)";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "PreSequencer";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "PostSequencer";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AutoStart";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "CATDSDevice";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "HugePages";
	prop_desc = "";
	prop_def  = "false";
	vect_data.clear();
	vect_data.push_back("false");
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "SharedMemoryID";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "EnergyDevice";
	prop_desc = "Device;attribute";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "EnergyRanges";
	prop_desc = "Range in which the energy is at every change is written to magnets and simulators.\nMinEnergy1;MaxEnergy1\nMinEnergy2;MaxEnergy2";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "HLSDevice";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "SourcePointLabels";
	prop_desc = "X_S1,Y_S1,XP_S1,YP_S1\nX_S2,Y_S2,XP_S2,YP_S2\n......";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "EnergyLoopPeriodSec";
	prop_desc = "Period in sec. of checking energy change and updating simulator and magnet energies";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "SyncLoopPeriodSec";
	prop_desc = "Period in sec. of synching CAT simulato rwith HLS in source points by modifying CAT source point gain and offset";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "MIMOFBList";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "PsDelayedStartList";
	prop_desc = "matched string; delay in seconds\nEx: *Q2_s*;3";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "OrbitDelayedStartList";
	prop_desc = "delay in seconds for horizontal plane\ndelay in second for vertical plane";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "TuneHorDevice";
	prop_desc = "tunedevice;tune_attribute";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "TuneVerDevice";
	prop_desc = "tunedevice;tune_attribute";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
}

//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::write_class_property()
 * Description:  Set class description fields as property in database
 */
//--------------------------------------------------------
void CATSyncClass::write_class_property()
{
//	First time, check if database used
if (Tango::Util::_UseDb == false)
	return;

Tango::DbData	data;
std::string	classname = get_name();
std::string	header;

//	Put title
Tango::DbDatum	title("ProjectTitle");
std::string	str_title("CATSync");
title << str_title;
data.push_back(title);

//	Put Description
Tango::DbDatum	description("Description");
std::vector<std::string>	str_desc;
str_desc.push_back("Tango server used to synchronize CAT simulator with field");
description << str_desc;
data.push_back(description);

//  Put inheritance
Tango::DbDatum	inher_datum("InheritedFrom");
std::vector<std::string> inheritance;
inheritance.push_back("TANGO_BASE_CLASS");
inher_datum << inheritance;
data.push_back(inher_datum);

//	Call database and and values
get_db_class()->put_property(data);
}

//===================================================================
//	Factory methods
//===================================================================

//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::device_factory()
 * Description:  Create the device object(s)
 *                and store them in the device list
 */
//--------------------------------------------------------
void CATSyncClass::device_factory(const Tango::DevVarStringArray *devlist_ptr)
{
/*----- PROTECTED REGION ID(CATSyncClass::device_factory_before) ENABLED START -----*/
/* clang-format on */
//	Add your own code
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSyncClass::device_factory_before

//	Create devices and add it into the device list
for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
{
	TANGO_LOG_DEBUG << "Device name : " << (*devlist_ptr)[i].in() << std::endl;
	device_list.push_back(new CATSync(this, (*devlist_ptr)[i]));
}

//	Manage dynamic attributes if any
erase_dynamic_attributes(devlist_ptr, get_class_attr()->get_attr_list());

//	Export devices to the outside world
for (unsigned long i=1 ; i<=devlist_ptr->length() ; i++)
{
	//	Add dynamic attributes if any
	CATSync *dev = static_cast<CATSync *>(device_list[device_list.size()-i]);
	dev->add_dynamic_attributes();

	//	Check before if database used.
	if ((Tango::Util::_UseDb == true) && (Tango::Util::_FileDb == false))
		export_device(dev);
	else
		export_device(dev, dev->get_name().c_str());
}

/*----- PROTECTED REGION ID(CATSyncClass::device_factory_after) ENABLED START -----*/
/* clang-format on */
//	Add your own code
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSyncClass::device_factory_after
}
//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::attribute_factory()
 * Description:  Create the attribute object(s)
 *                and store them in the attribute list
 */
//--------------------------------------------------------
void CATSyncClass::attribute_factory(std::vector<Tango::Attr *> &att_list)
{
	/*----- PROTECTED REGION ID(CATSyncClass::attribute_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::attribute_factory_before
	//	Attribute : EnableCATEnergyUpdate
	EnableCATEnergyUpdateAttrib	*enablecatenergyupdate = new EnableCATEnergyUpdateAttrib();
	Tango::UserDefaultAttrProp	enablecatenergyupdate_prop;
	//	description	not set for EnableCATEnergyUpdate
	//	label	not set for EnableCATEnergyUpdate
	//	unit	not set for EnableCATEnergyUpdate
	//	standard_unit	not set for EnableCATEnergyUpdate
	//	display_unit	not set for EnableCATEnergyUpdate
	//	format	not set for EnableCATEnergyUpdate
	//	max_value	not set for EnableCATEnergyUpdate
	//	min_value	not set for EnableCATEnergyUpdate
	//	max_alarm	not set for EnableCATEnergyUpdate
	//	min_alarm	not set for EnableCATEnergyUpdate
	//	max_warning	not set for EnableCATEnergyUpdate
	//	min_warning	not set for EnableCATEnergyUpdate
	//	delta_t	not set for EnableCATEnergyUpdate
	//	delta_val	not set for EnableCATEnergyUpdate
	enablecatenergyupdate->set_default_properties(enablecatenergyupdate_prop);
	//	Not Polled
	enablecatenergyupdate->set_disp_level(Tango::OPERATOR);
	enablecatenergyupdate->set_memorized();
	enablecatenergyupdate->set_memorized_init(true);
	att_list.push_back(enablecatenergyupdate);

	//	Attribute : EnableCATSourcePointUpdate
	EnableCATSourcePointUpdateAttrib	*enablecatsourcepointupdate = new EnableCATSourcePointUpdateAttrib();
	Tango::UserDefaultAttrProp	enablecatsourcepointupdate_prop;
	//	description	not set for EnableCATSourcePointUpdate
	//	label	not set for EnableCATSourcePointUpdate
	//	unit	not set for EnableCATSourcePointUpdate
	//	standard_unit	not set for EnableCATSourcePointUpdate
	//	display_unit	not set for EnableCATSourcePointUpdate
	//	format	not set for EnableCATSourcePointUpdate
	//	max_value	not set for EnableCATSourcePointUpdate
	//	min_value	not set for EnableCATSourcePointUpdate
	//	max_alarm	not set for EnableCATSourcePointUpdate
	//	min_alarm	not set for EnableCATSourcePointUpdate
	//	max_warning	not set for EnableCATSourcePointUpdate
	//	min_warning	not set for EnableCATSourcePointUpdate
	//	delta_t	not set for EnableCATSourcePointUpdate
	//	delta_val	not set for EnableCATSourcePointUpdate
	enablecatsourcepointupdate->set_default_properties(enablecatsourcepointupdate_prop);
	//	Not Polled
	enablecatsourcepointupdate->set_disp_level(Tango::OPERATOR);
	enablecatsourcepointupdate->set_memorized();
	enablecatsourcepointupdate->set_memorized_init(true);
	att_list.push_back(enablecatsourcepointupdate);

	//	Attribute : LFBCATTimeFactor
	LFBCATTimeFactorAttrib	*lfbcattimefactor = new LFBCATTimeFactorAttrib();
	Tango::UserDefaultAttrProp	lfbcattimefactor_prop;
	//	description	not set for LFBCATTimeFactor
	//	label	not set for LFBCATTimeFactor
	//	unit	not set for LFBCATTimeFactor
	//	standard_unit	not set for LFBCATTimeFactor
	//	display_unit	not set for LFBCATTimeFactor
	//	format	not set for LFBCATTimeFactor
	//	max_value	not set for LFBCATTimeFactor
	//	min_value	not set for LFBCATTimeFactor
	//	max_alarm	not set for LFBCATTimeFactor
	//	min_alarm	not set for LFBCATTimeFactor
	//	max_warning	not set for LFBCATTimeFactor
	//	min_warning	not set for LFBCATTimeFactor
	//	delta_t	not set for LFBCATTimeFactor
	//	delta_val	not set for LFBCATTimeFactor
	lfbcattimefactor->set_default_properties(lfbcattimefactor_prop);
	//	Not Polled
	lfbcattimefactor->set_disp_level(Tango::OPERATOR);
	lfbcattimefactor->set_memorized();
	lfbcattimefactor->set_memorized_init(true);
	att_list.push_back(lfbcattimefactor);

	//	Attribute : LFBHLSTimeFactor
	LFBHLSTimeFactorAttrib	*lfbhlstimefactor = new LFBHLSTimeFactorAttrib();
	Tango::UserDefaultAttrProp	lfbhlstimefactor_prop;
	//	description	not set for LFBHLSTimeFactor
	//	label	not set for LFBHLSTimeFactor
	//	unit	not set for LFBHLSTimeFactor
	//	standard_unit	not set for LFBHLSTimeFactor
	//	display_unit	not set for LFBHLSTimeFactor
	//	format	not set for LFBHLSTimeFactor
	//	max_value	not set for LFBHLSTimeFactor
	//	min_value	not set for LFBHLSTimeFactor
	//	max_alarm	not set for LFBHLSTimeFactor
	//	min_alarm	not set for LFBHLSTimeFactor
	//	max_warning	not set for LFBHLSTimeFactor
	//	min_warning	not set for LFBHLSTimeFactor
	//	delta_t	not set for LFBHLSTimeFactor
	//	delta_val	not set for LFBHLSTimeFactor
	lfbhlstimefactor->set_default_properties(lfbhlstimefactor_prop);
	//	Not Polled
	lfbhlstimefactor->set_disp_level(Tango::OPERATOR);
	lfbhlstimefactor->set_memorized();
	lfbhlstimefactor->set_memorized_init(true);
	att_list.push_back(lfbhlstimefactor);

	//	Attribute : EnableCATMatchOrbit
	EnableCATMatchOrbitAttrib	*enablecatmatchorbit = new EnableCATMatchOrbitAttrib();
	Tango::UserDefaultAttrProp	enablecatmatchorbit_prop;
	//	description	not set for EnableCATMatchOrbit
	//	label	not set for EnableCATMatchOrbit
	//	unit	not set for EnableCATMatchOrbit
	//	standard_unit	not set for EnableCATMatchOrbit
	//	display_unit	not set for EnableCATMatchOrbit
	//	format	not set for EnableCATMatchOrbit
	//	max_value	not set for EnableCATMatchOrbit
	//	min_value	not set for EnableCATMatchOrbit
	//	max_alarm	not set for EnableCATMatchOrbit
	//	min_alarm	not set for EnableCATMatchOrbit
	//	max_warning	not set for EnableCATMatchOrbit
	//	min_warning	not set for EnableCATMatchOrbit
	//	delta_t	not set for EnableCATMatchOrbit
	//	delta_val	not set for EnableCATMatchOrbit
	enablecatmatchorbit->set_default_properties(enablecatmatchorbit_prop);
	//	Not Polled
	enablecatmatchorbit->set_disp_level(Tango::OPERATOR);
	enablecatmatchorbit->set_memorized();
	enablecatmatchorbit->set_memorized_init(true);
	att_list.push_back(enablecatmatchorbit);

	//	Attribute : EnableCATSyncMagnets
	EnableCATSyncMagnetsAttrib	*enablecatsyncmagnets = new EnableCATSyncMagnetsAttrib();
	Tango::UserDefaultAttrProp	enablecatsyncmagnets_prop;
	//	description	not set for EnableCATSyncMagnets
	//	label	not set for EnableCATSyncMagnets
	//	unit	not set for EnableCATSyncMagnets
	//	standard_unit	not set for EnableCATSyncMagnets
	//	display_unit	not set for EnableCATSyncMagnets
	//	format	not set for EnableCATSyncMagnets
	//	max_value	not set for EnableCATSyncMagnets
	//	min_value	not set for EnableCATSyncMagnets
	//	max_alarm	not set for EnableCATSyncMagnets
	//	min_alarm	not set for EnableCATSyncMagnets
	//	max_warning	not set for EnableCATSyncMagnets
	//	min_warning	not set for EnableCATSyncMagnets
	//	delta_t	not set for EnableCATSyncMagnets
	//	delta_val	not set for EnableCATSyncMagnets
	enablecatsyncmagnets->set_default_properties(enablecatsyncmagnets_prop);
	//	Not Polled
	enablecatsyncmagnets->set_disp_level(Tango::OPERATOR);
	enablecatsyncmagnets->set_memorized();
	enablecatsyncmagnets->set_memorized_init(true);
	att_list.push_back(enablecatsyncmagnets);

	//	Attribute : CATState
	CATStateAttrib	*catstate = new CATStateAttrib();
	Tango::UserDefaultAttrProp	catstate_prop;
	//	description	not set for CATState
	//	label	not set for CATState
	//	unit	not set for CATState
	//	standard_unit	not set for CATState
	//	display_unit	not set for CATState
	//	format	not set for CATState
	//	max_value	not set for CATState
	//	min_value	not set for CATState
	//	max_alarm	not set for CATState
	//	min_alarm	not set for CATState
	//	max_warning	not set for CATState
	//	min_warning	not set for CATState
	//	delta_t	not set for CATState
	//	delta_val	not set for CATState
	catstate->set_default_properties(catstate_prop);
	//	Not Polled
	catstate->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(catstate);

	//	Attribute : HLSState
	HLSStateAttrib	*hlsstate = new HLSStateAttrib();
	Tango::UserDefaultAttrProp	hlsstate_prop;
	//	description	not set for HLSState
	//	label	not set for HLSState
	//	unit	not set for HLSState
	//	standard_unit	not set for HLSState
	//	display_unit	not set for HLSState
	//	format	not set for HLSState
	//	max_value	not set for HLSState
	//	min_value	not set for HLSState
	//	max_alarm	not set for HLSState
	//	min_alarm	not set for HLSState
	//	max_warning	not set for HLSState
	//	min_warning	not set for HLSState
	//	delta_t	not set for HLSState
	//	delta_val	not set for HLSState
	hlsstate->set_default_properties(hlsstate_prop);
	//	Not Polled
	hlsstate->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(hlsstate);

	//	Attribute : SyncGain
	SyncGainAttrib	*syncgain = new SyncGainAttrib();
	Tango::UserDefaultAttrProp	syncgain_prop;
	//	description	not set for SyncGain
	//	label	not set for SyncGain
	//	unit	not set for SyncGain
	//	standard_unit	not set for SyncGain
	//	display_unit	not set for SyncGain
	syncgain_prop.set_format("%.5f");
	syncgain_prop.set_max_value("1");
	syncgain_prop.set_min_value("0");
	//	max_alarm	not set for SyncGain
	//	min_alarm	not set for SyncGain
	//	max_warning	not set for SyncGain
	//	min_warning	not set for SyncGain
	//	delta_t	not set for SyncGain
	//	delta_val	not set for SyncGain
	syncgain->set_default_properties(syncgain_prop);
	//	Not Polled
	syncgain->set_disp_level(Tango::OPERATOR);
	syncgain->set_memorized();
	syncgain->set_memorized_init(true);
	att_list.push_back(syncgain);

	//	Attribute : CATTuneHor
	CATTuneHorAttrib	*cattunehor = new CATTuneHorAttrib();
	Tango::UserDefaultAttrProp	cattunehor_prop;
	//	description	not set for CATTuneHor
	//	label	not set for CATTuneHor
	//	unit	not set for CATTuneHor
	//	standard_unit	not set for CATTuneHor
	//	display_unit	not set for CATTuneHor
	cattunehor_prop.set_format("%.5f");
	cattunehor_prop.set_max_value("1");
	cattunehor_prop.set_min_value("0");
	//	max_alarm	not set for CATTuneHor
	//	min_alarm	not set for CATTuneHor
	//	max_warning	not set for CATTuneHor
	//	min_warning	not set for CATTuneHor
	//	delta_t	not set for CATTuneHor
	//	delta_val	not set for CATTuneHor
	cattunehor->set_default_properties(cattunehor_prop);
	//	Not Polled
	cattunehor->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(cattunehor);

	//	Attribute : CATTuneVer
	CATTuneVerAttrib	*cattunever = new CATTuneVerAttrib();
	Tango::UserDefaultAttrProp	cattunever_prop;
	//	description	not set for CATTuneVer
	//	label	not set for CATTuneVer
	//	unit	not set for CATTuneVer
	//	standard_unit	not set for CATTuneVer
	//	display_unit	not set for CATTuneVer
	cattunever_prop.set_format("%.5f");
	cattunever_prop.set_max_value("1");
	cattunever_prop.set_min_value("0");
	//	max_alarm	not set for CATTuneVer
	//	min_alarm	not set for CATTuneVer
	//	max_warning	not set for CATTuneVer
	//	min_warning	not set for CATTuneVer
	//	delta_t	not set for CATTuneVer
	//	delta_val	not set for CATTuneVer
	cattunever->set_default_properties(cattunever_prop);
	//	Not Polled
	cattunever->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(cattunever);

	//	Attribute : MatchThreadErrMsg
	MatchThreadErrMsgAttrib	*matchthreaderrmsg = new MatchThreadErrMsgAttrib();
	Tango::UserDefaultAttrProp	matchthreaderrmsg_prop;
	//	description	not set for MatchThreadErrMsg
	//	label	not set for MatchThreadErrMsg
	//	unit	not set for MatchThreadErrMsg
	//	standard_unit	not set for MatchThreadErrMsg
	//	display_unit	not set for MatchThreadErrMsg
	//	format	not set for MatchThreadErrMsg
	//	max_value	not set for MatchThreadErrMsg
	//	min_value	not set for MatchThreadErrMsg
	//	max_alarm	not set for MatchThreadErrMsg
	//	min_alarm	not set for MatchThreadErrMsg
	//	max_warning	not set for MatchThreadErrMsg
	//	min_warning	not set for MatchThreadErrMsg
	//	delta_t	not set for MatchThreadErrMsg
	//	delta_val	not set for MatchThreadErrMsg
	matchthreaderrmsg->set_default_properties(matchthreaderrmsg_prop);
	//	Not Polled
	matchthreaderrmsg->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(matchthreaderrmsg);

	//	Attribute : SyncThreadErrMsg
	SyncThreadErrMsgAttrib	*syncthreaderrmsg = new SyncThreadErrMsgAttrib();
	Tango::UserDefaultAttrProp	syncthreaderrmsg_prop;
	//	description	not set for SyncThreadErrMsg
	//	label	not set for SyncThreadErrMsg
	//	unit	not set for SyncThreadErrMsg
	//	standard_unit	not set for SyncThreadErrMsg
	//	display_unit	not set for SyncThreadErrMsg
	//	format	not set for SyncThreadErrMsg
	//	max_value	not set for SyncThreadErrMsg
	//	min_value	not set for SyncThreadErrMsg
	//	max_alarm	not set for SyncThreadErrMsg
	//	min_alarm	not set for SyncThreadErrMsg
	//	max_warning	not set for SyncThreadErrMsg
	//	min_warning	not set for SyncThreadErrMsg
	//	delta_t	not set for SyncThreadErrMsg
	//	delta_val	not set for SyncThreadErrMsg
	syncthreaderrmsg->set_default_properties(syncthreaderrmsg_prop);
	//	Not Polled
	syncthreaderrmsg->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(syncthreaderrmsg);

	//	Attribute : MatchThreadErrCnt
	MatchThreadErrCntAttrib	*matchthreaderrcnt = new MatchThreadErrCntAttrib();
	Tango::UserDefaultAttrProp	matchthreaderrcnt_prop;
	//	description	not set for MatchThreadErrCnt
	//	label	not set for MatchThreadErrCnt
	//	unit	not set for MatchThreadErrCnt
	//	standard_unit	not set for MatchThreadErrCnt
	//	display_unit	not set for MatchThreadErrCnt
	//	format	not set for MatchThreadErrCnt
	//	max_value	not set for MatchThreadErrCnt
	//	min_value	not set for MatchThreadErrCnt
	//	max_alarm	not set for MatchThreadErrCnt
	//	min_alarm	not set for MatchThreadErrCnt
	//	max_warning	not set for MatchThreadErrCnt
	//	min_warning	not set for MatchThreadErrCnt
	//	delta_t	not set for MatchThreadErrCnt
	//	delta_val	not set for MatchThreadErrCnt
	matchthreaderrcnt->set_default_properties(matchthreaderrcnt_prop);
	//	Not Polled
	matchthreaderrcnt->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(matchthreaderrcnt);

	//	Attribute : SyncThreadErrCnt
	SyncThreadErrCntAttrib	*syncthreaderrcnt = new SyncThreadErrCntAttrib();
	Tango::UserDefaultAttrProp	syncthreaderrcnt_prop;
	//	description	not set for SyncThreadErrCnt
	//	label	not set for SyncThreadErrCnt
	//	unit	not set for SyncThreadErrCnt
	//	standard_unit	not set for SyncThreadErrCnt
	//	display_unit	not set for SyncThreadErrCnt
	//	format	not set for SyncThreadErrCnt
	//	max_value	not set for SyncThreadErrCnt
	//	min_value	not set for SyncThreadErrCnt
	//	max_alarm	not set for SyncThreadErrCnt
	//	min_alarm	not set for SyncThreadErrCnt
	//	max_warning	not set for SyncThreadErrCnt
	//	min_warning	not set for SyncThreadErrCnt
	//	delta_t	not set for SyncThreadErrCnt
	//	delta_val	not set for SyncThreadErrCnt
	syncthreaderrcnt->set_default_properties(syncthreaderrcnt_prop);
	//	Not Polled
	syncthreaderrcnt->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(syncthreaderrcnt);

	//	Attribute : MatchThreadCnt
	MatchThreadCntAttrib	*matchthreadcnt = new MatchThreadCntAttrib();
	Tango::UserDefaultAttrProp	matchthreadcnt_prop;
	//	description	not set for MatchThreadCnt
	//	label	not set for MatchThreadCnt
	//	unit	not set for MatchThreadCnt
	//	standard_unit	not set for MatchThreadCnt
	//	display_unit	not set for MatchThreadCnt
	//	format	not set for MatchThreadCnt
	//	max_value	not set for MatchThreadCnt
	//	min_value	not set for MatchThreadCnt
	//	max_alarm	not set for MatchThreadCnt
	//	min_alarm	not set for MatchThreadCnt
	//	max_warning	not set for MatchThreadCnt
	//	min_warning	not set for MatchThreadCnt
	//	delta_t	not set for MatchThreadCnt
	//	delta_val	not set for MatchThreadCnt
	matchthreadcnt->set_default_properties(matchthreadcnt_prop);
	//	Not Polled
	matchthreadcnt->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(matchthreadcnt);

	//	Attribute : SyncThreadCnt
	SyncThreadCntAttrib	*syncthreadcnt = new SyncThreadCntAttrib();
	Tango::UserDefaultAttrProp	syncthreadcnt_prop;
	//	description	not set for SyncThreadCnt
	//	label	not set for SyncThreadCnt
	//	unit	not set for SyncThreadCnt
	//	standard_unit	not set for SyncThreadCnt
	//	display_unit	not set for SyncThreadCnt
	//	format	not set for SyncThreadCnt
	//	max_value	not set for SyncThreadCnt
	//	min_value	not set for SyncThreadCnt
	//	max_alarm	not set for SyncThreadCnt
	//	min_alarm	not set for SyncThreadCnt
	//	max_warning	not set for SyncThreadCnt
	//	min_warning	not set for SyncThreadCnt
	//	delta_t	not set for SyncThreadCnt
	//	delta_val	not set for SyncThreadCnt
	syncthreadcnt->set_default_properties(syncthreadcnt_prop);
	//	Not Polled
	syncthreadcnt->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(syncthreadcnt);


	//	Create a list of static attributes
	create_static_attribute_list(get_class_attr()->get_attr_list());
	/*----- PROTECTED REGION ID(CATSyncClass::attribute_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::attribute_factory_after
}
//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::pipe_factory()
 * Description:  Create the pipe object(s)
 *                and store them in the pipe list
 */
//--------------------------------------------------------
void CATSyncClass::pipe_factory()
{
	/*----- PROTECTED REGION ID(CATSyncClass::pipe_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::pipe_factory_before
	/*----- PROTECTED REGION ID(CATSyncClass::pipe_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::pipe_factory_after
}
//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::command_factory()
 * Description:  Create the command object(s)
 *                and store them in the command list
 */
//--------------------------------------------------------
void CATSyncClass::command_factory()
{
	/*----- PROTECTED REGION ID(CATSyncClass::command_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::command_factory_before


	//	Command Start
	StartClass	*pStartCmd =
		new StartClass("Start",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pStartCmd);

	//	Command Stop
	StopClass	*pStopCmd =
		new StopClass("Stop",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pStopCmd);

	//	Command GotoEnergy
	GotoEnergyClass	*pGotoEnergyCmd =
		new GotoEnergyClass("GotoEnergy",
			Tango::DEV_DOUBLE, Tango::DEV_VOID,
			"Energy (GeV)",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGotoEnergyCmd);

	//	Command UpdateEnergy
	UpdateEnergyClass	*pUpdateEnergyCmd =
		new UpdateEnergyClass("UpdateEnergy",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pUpdateEnergyCmd);

	//	Command SyncSourcePointsToHLS
	SyncSourcePointsToHLSClass	*pSyncSourcePointsToHLSCmd =
		new SyncSourcePointsToHLSClass("SyncSourcePointsToHLS",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pSyncSourcePointsToHLSCmd);

	//	Command SwitchLFBCAT
	SwitchLFBCATClass	*pSwitchLFBCATCmd =
		new SwitchLFBCATClass("SwitchLFBCAT",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pSwitchLFBCATCmd);

	//	Command SwitchLFBHLS
	SwitchLFBHLSClass	*pSwitchLFBHLSCmd =
		new SwitchLFBHLSClass("SwitchLFBHLS",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pSwitchLFBHLSCmd);

	//	Command ConvertSourcePointsMM
	ConvertSourcePointsMMClass	*pConvertSourcePointsMMCmd =
		new ConvertSourcePointsMMClass("ConvertSourcePointsMM",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pConvertSourcePointsMMCmd);

	//	Command ResetSourcePointOffsets
	ResetSourcePointOffsetsClass	*pResetSourcePointOffsetsCmd =
		new ResetSourcePointOffsetsClass("ResetSourcePointOffsets",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pResetSourcePointOffsetsCmd);

	/*----- PROTECTED REGION ID(CATSyncClass::command_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::command_factory_after
}

//===================================================================
//	Dynamic attributes related methods
//===================================================================

//--------------------------------------------------------
/**
 * method : 		CATSyncClass::create_static_attribute_list
 * description : 	Create the a list of static attributes
 *
 * @param	att_list	the created attribute list
 */
//--------------------------------------------------------
void CATSyncClass::create_static_attribute_list(std::vector<Tango::Attr *> &att_list)
{
	for (unsigned long i=0 ; i<att_list.size() ; i++)
	{
		std::string att_name(att_list[i]->get_name());
		transform(att_name.begin(), att_name.end(), att_name.begin(), ::tolower);
		defaultAttList.push_back(att_name);
	}

	TANGO_LOG_INFO << defaultAttList.size() << " attributes in default list" << std::endl;

	/*----- PROTECTED REGION ID(CATSyncClass::create_static_att_list) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::create_static_att_list
}


//--------------------------------------------------------
/**
 * method : 		CATSyncClass::erase_dynamic_attributes
 * description : 	delete the dynamic attributes if any.
 *
 * @param	devlist_ptr	the device list pointer
 * @param	list of all attributes
 */
//--------------------------------------------------------
void CATSyncClass::erase_dynamic_attributes(const Tango::DevVarStringArray *devlist_ptr, std::vector<Tango::Attr *> &att_list)
{
	Tango::Util *tg = Tango::Util::instance();

	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		Tango::DeviceImpl *dev_impl = tg->get_device_by_name(((std::string)(*devlist_ptr)[i]).c_str());
		CATSync *dev = static_cast<CATSync *> (dev_impl);

		std::vector<Tango::Attribute *> &dev_att_list = dev->get_device_attr()->get_attribute_list();
		std::vector<Tango::Attribute *>::iterator ite_att;
		for (ite_att=dev_att_list.begin() ; ite_att != dev_att_list.end() ; ++ite_att)
		{
			std::string att_name((*ite_att)->get_name_lower());
			if ((att_name == "state") || (att_name == "status"))
				continue;
			std::vector<std::string>::iterator ite_str = find(defaultAttList.begin(), defaultAttList.end(), att_name);
			if (ite_str == defaultAttList.end())
			{
				TANGO_LOG_INFO << att_name << " is a UNWANTED dynamic attribute for device " << (*devlist_ptr)[i] << std::endl;
				Tango::Attribute &att = dev->get_device_attr()->get_attr_by_name(att_name.c_str());
				dev->remove_attribute(att_list[att.get_attr_idx()], true, false);
				--ite_att;
			}
		}
	}
	/*----- PROTECTED REGION ID(CATSyncClass::erase_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CATSyncClass::erase_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Method      : CATSyncClass::get_attr_object_by_name()
 * Description:  returns Tango::Attr * object found by name
 */
//--------------------------------------------------------
Tango::Attr *CATSyncClass::get_attr_object_by_name(std::vector<Tango::Attr *> &att_list, std::string attname)
{
	std::vector<Tango::Attr *>::iterator it;
	for (it=att_list.begin() ; it<att_list.end() ; ++it)
		if ((*it)->get_name()==attname)
			return (*it);
	//	Attr does not exist
	return NULL;
}


/*----- PROTECTED REGION ID(CATSyncClass::Additional Methods) ENABLED START -----*/
/* clang-format on */
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CATSyncClass::Additional Methods
} //	namespace
