//
// file :         catsyncthread.cpp
//
// description :  bunchnumber thread
//
// project :      TANGO Device Server
//
// $Author: giulio $ 

// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AR
//              34012 Basovizza, Trieste ITALY
//

#include <omnithread.h>
#include "CATSync.h"
#include "catsyncthread.h"

namespace CATSync_ns
{

/* 
 *	catsyncthread::run()
 *	Run  
 */		
void catsyncthread::run(void *) 
{
	struct timespec now;

	if (mycatsync->ps_magnet_sync[thread_index].magnetdevice.size() == 0)
		return;

	mycatsync->run_catsync_flag[thread_index] = true;

	while (mycatsync->run_sequencer_flag[CATSYNC_PRESEQUENCER_IDX]) {
		omni_thread::sleep(0,100000000);
	}

	for (int i = 0; i < mycatsync->ps_magnet_sync[thread_index].magnetdevice.size(); i++) {
		mycatsync->ps_magnet_sync[thread_index].old_value.push_back(NAN);
		mycatsync->ps_magnet_sync[thread_index].value.push_back(NAN);
	}

	clock_gettime(CLOCK_REALTIME, &mycatsync->init_cat_time);

	//mycatsync->init_cat_time.tv_sec -= 100;

	int cnt = 0;

	INFO_STREAM << __FUNCTION__ << ", start sync thread " << thread_index << endl;

	while (!mycatsync->abort_catsync_flag[thread_index]) {

		clock_gettime(CLOCK_REALTIME, &now);

		if (*mycatsync->attr_EnableCATSyncMagnets_read == false) {
			omni_thread::sleep(1,0);
			continue;
		}

		INFO_STREAM << __FUNCTION__ << ", loop sync thread " << thread_index << endl;

		for (int i = 0; i < mycatsync->ps_magnet_sync[thread_index].magnetdevice.size(); i++) {
			double rd_value = NAN, rd_rr_value, old_rd_value;

			if (mycatsync->abort_catsync_flag[thread_index])
				break;

			// delay reading
			if (mycatsync->ps_magnet_sync[thread_index].start_time[i].tv_sec > now.tv_sec) {
				continue;
			}

			try {
				Tango::DeviceAttribute Dout;
				vector <double> value;
				vector <string> value_str;

				Dout = mycatsync->ps_magnet_sync[thread_index].magnetdevice[i]->read_attribute(
					mycatsync->ps_magnet_sync[thread_index].magnet_attrname[i]);
				// cast numeric data into double
				extract_values(&Dout, value, value_str);
				if ((value.size() + 1) >= mycatsync->ps_magnet_sync[thread_index].vec_idx[i]) {
					rd_value = value[mycatsync->ps_magnet_sync[thread_index].vec_idx[i]];
					/*rd_rr_value = value[0];
					if (rd_value != old_rd_value) {
						double max_abs_value = (fabs(rd_value) > fabs(old_rd_value)) ? fabs(rd_value) : fabs(old_rd_value);
						double difference = fabs(rd_value - rd_rr_value);
						if (difference > 0.05 * max_bas_value)
							rd_value = rd_rr_value;
					}*/
				}

			}
			catch (Tango::DevFailed &e) {
				rd_value = NAN;
				mycatsync->ps_magnet_sync[thread_index].old_value[i] = NAN;
				TangoSys_MemStream error_stream;
				error_stream << e.errors[0].desc << ", error reading " << 
					mycatsync->ps_magnet_sync[thread_index].magnet_devicename[i] << ", " << 
					mycatsync->ps_magnet_sync[thread_index].magnet_attrname[i];
				INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
				strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
				(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;
			}
			catch(...) {
				rd_value = NAN;
				mycatsync->ps_magnet_sync[thread_index].old_value[i] = NAN;
				TangoSys_MemStream error_stream;
				error_stream << "generic error reading " << 
					mycatsync->ps_magnet_sync[thread_index].magnet_devicename[i] << ", " << 
					mycatsync->ps_magnet_sync[thread_index].magnet_attrname[i];
				INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;	
				strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
				(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;
			}

			if (!isnan(rd_value)) {
				Tango::DeviceAttribute *Din;

				rd_value *= mycatsync->ps_magnet_sync[thread_index].conv_factor[i];
				mycatsync->ps_magnet_sync[thread_index].old_value[i] = mycatsync->ps_magnet_sync[thread_index].value[i];
				old_rd_value = mycatsync->ps_magnet_sync[thread_index].value[i];
				mycatsync->ps_magnet_sync[thread_index].value[i] = rd_value;			

				

				INFO_STREAM << __FUNCTION__ << ", checking " << mycatsync->ps_magnet_sync[thread_index].sim_attrname[i] <<
					", old value " << old_rd_value << ", new value " << rd_value << endl;

				//if ((mycatsync->ps_magnet_sync[thread_index].value[i] != mycatsync->ps_magnet_sync[thread_index].old_value[i]) ||
				//	((cnt % 100) == 0) ) {
				if (1) {
					try {
						vector <double> data;
						mycatsync->catdevice->read_attribute(mycatsync->ps_magnet_sync[thread_index].sim_attrname[i].c_str()) >> data;

						//printf("data 0=%f 1=%f\n",data[0], data[1]);

						vector <double> vval;
						vector <string> ss, value_str;

						double wr_data = data[0] + (rd_value - data[0]) * *mycatsync->attr_SyncGain_read;

						vval.push_back(wr_data);
						Tango::DeviceAttribute Dout;					

						Tango::AttributeInfoEx attr_info = mycatsync->catdevice->get_attribute_config(
							mycatsync->ps_magnet_sync[thread_index].sim_attrname[i]);
						
						
						INFO_STREAM << __FUNCTION__ << ", write " << mycatsync->ps_magnet_sync[thread_index].sim_attrname[i] << "" << wr_data << endl;

						Din = insert_values(attr_info, mycatsync->ps_magnet_sync[thread_index].sim_attrname[i], vval, ss);
						// write simulator attribute
						mycatsync->catdevice->write_attribute(*Din);

						if (Din) {
							delete Din; Din = 0;
						}
						(mycatsync->attr_SyncThreadCnt_read[thread_index])++;

					}
					catch (Tango::DevFailed &e) {
						if (Din) {
							delete Din; Din = 0;
						}
						mycatsync->ps_magnet_sync[thread_index].old_value[i] = NAN;
						TangoSys_MemStream error_stream;
						error_stream << e.errors[0].desc << ", error setting " << mycatsync->cATDSDevice << ", " << 
							mycatsync->ps_magnet_sync[thread_index].sim_attrname[i];
						INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
						strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
						(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;
					}
					catch(...) {
						if (Din) {
							delete Din; Din = 0;
						}
						mycatsync->ps_magnet_sync[thread_index].old_value[i] = NAN;
						TangoSys_MemStream error_stream;
						error_stream << "generic error reading " << 
							mycatsync->ps_magnet_sync[thread_index].sim_attrname[i];
						INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
						strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
						(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;		
					}	
				}
				else {
					// skipped
				}	
			}

			

		} // for


		// update energies of magnets and CAT simulator
		if (abs(now.tv_sec - mycatsync->energy_update_time[thread_index].tv_sec) > mycatsync->energyLoopPeriodSec) {
			try {
				mycatsync->device_update_energy(thread_index);		
			}
			catch (Tango::DevFailed &e) {
				TangoSys_MemStream error_stream;
				error_stream << e.errors[0].desc << ", error updating energy";
				INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
				strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
				(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;
			}
			catch(...) {
				TangoSys_MemStream error_stream;
				error_stream << "generic error, error updating energy";
				INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
				strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
				(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;		
			}
			clock_gettime(CLOCK_REALTIME, &mycatsync->energy_update_time[thread_index]);
		}

		// sync the CAT source points with HLS ones by changing errgain and erroffset of CAT sourcepoints
		if ((thread_index == (CATSYNC_MAX_THREADS-1)) &&  *mycatsync->attr_EnableCATSourcePointUpdate_read) {
			if (abs(now.tv_sec - mycatsync->source_points_update_time.tv_sec) > mycatsync->syncLoopPeriodSec) {
				try {
					mycatsync->synchronize_attributes();		
				}
				catch (Tango::DevFailed &e) {
					TangoSys_MemStream error_stream;
					error_stream << e.errors[0].desc << ", error updating CAT source points";
					INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
					strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
					(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;
				}
				catch(...) {
					TangoSys_MemStream error_stream;
					error_stream << "generic error, error updating CAT source points";
					INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
					strncpy(mycatsync->attr_SyncThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
					(mycatsync->attr_SyncThreadErrCnt_read[thread_index])++;		
				}
				clock_gettime(CLOCK_REALTIME, &mycatsync->source_points_update_time);
			}
		}

		if (thread_index == (CATSYNC_MAX_THREADS-1)) {
			Tango::DevState cat_state;
			try {
				mycatsync->catdevice->read_attribute("State") >> cat_state;
				if (cat_state == Tango::FAULT) {
					*mycatsync->attr_SyncGain_read = 0.0001;
					mycatsync->catdevice->command_inout("Init");
					clock_gettime(CLOCK_REALTIME, &mycatsync->init_cat_time);
					omni_thread::sleep(10,0); // 10ms
				}
				if (now.tv_sec - mycatsync->init_cat_time.tv_sec > 60) {
					*mycatsync->attr_SyncGain_read = 1;
				}
				else if (now.tv_sec - mycatsync->init_cat_time.tv_sec > 40) {
					*mycatsync->attr_SyncGain_read = 0.01;
				}			
				else if (now.tv_sec - mycatsync->init_cat_time.tv_sec > 20) {
					*mycatsync->attr_SyncGain_read = 0.001;
				}
				else {
					*mycatsync->attr_SyncGain_read = 0.0001;
				}
			}
			catch(...) {
				*mycatsync->attr_SyncGain_read = 0.0001;
				clock_gettime(CLOCK_REALTIME, &mycatsync->init_cat_time);
			}			
		}


		omni_thread::sleep(0,10000000); // 10ms
		cnt++;

	}

	// the thread with the lowest index start the post sequencer
	for (int i = 0; i < CATSYNC_MAX_THREADS; i++) {
		if ((mycatsync->abort_catsync_flag[i] == true) && (thread_index == i)) {
			mycatsync->start_sequencer(CATSYNC_POSTSEQUENCER_IDX);
			break;
		}
	}

	mycatsync->abort_catsync_flag[thread_index] = false;	
	mycatsync->run_catsync_flag[thread_index] = false;

}

}

