//
// file :         catsyncthread.cpp
//
// description :  bunchnumber thread
//
// project :      TANGO Device Server
//
// $Author: giulio $ 

// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AR
//              34012 Basovizza, Trieste ITALY
//

#include <omnithread.h>
#include "CATSync.h"
#include "matchthread.h"

namespace CATSync_ns
{

/* 
 *	matchthread::run()
 *	Run  
 */		
void matchthread::run(void *) 
{
	vector <string> sim_attrnames = {"matchfbx_ref_sensors","matchfby_ref_sensors"};
	vector <string> bpm_attrnames = {"XPosSa","ZPosSa"};
	mycatsync->run_match_orb_flag[thread_index] = true;
	float bpm_pos[200];
	int32_t slow_buf_idx[200];
	struct timespec now;
	// get current settings

	Tango::GroupReply::enable_exception(true);
	Tango::GroupAttrReplyList bpm_crl;

	while (mycatsync->run_sequencer_flag[CATSYNC_PRESEQUENCER_IDX]) {
		omni_thread::sleep(0,100000000);
	}

	INFO_STREAM << __FUNCTION__ << ", start match thread " << thread_index << endl;

	while (!mycatsync->abort_match_orb_flag[thread_index]) {

		Tango::DeviceAttribute *Din;
		vector <double> value;
		vector <string> value_str;
		bool err_flag = false;
		int bpm_size;

		clock_gettime(CLOCK_REALTIME, &now);

		if (*mycatsync->attr_EnableCATMatchOrbit_read == false) {
			omni_thread::sleep(1,0);
			continue;
		}

		// delay reading
		if (mycatsync->orbit_start_time[thread_index].tv_sec > now.tv_sec) {
			continue;
		}

		INFO_STREAM << __FUNCTION__ << ", run match thread " << thread_index << endl;

#ifdef DPDK
		for (int32_t i = 0; i < mycatsync->gof2_bpm_idx.size(); i++) {
			slow_buf_idx[i] = (mycatsync->gof2->bpm[mycatsync->gof2_bpm_idx[i]].slow_buf_idx - 1 + GOF2_BUF_SIZE) % GOF2_BUF_SIZE; // index inside the circular buffer
		}

		for (uint32_t i = 0; i < mycatsync->gof2_bpm_idx.size(); i++) {
			if (thread_index == 0)  // XPosSa
				bpm_pos[i] = mycatsync->gof2->bpm[mycatsync->gof2_bpm_idx[i]].slow_bpm_data[slow_buf_idx[i]].X;
			else if (thread_index == 1)
				bpm_pos[i] = mycatsync->gof2->bpm[mycatsync->gof2_bpm_idx[i]].slow_bpm_data[slow_buf_idx[i]].Y;

			bpm_pos[i] /= 1000;
			INFO_STREAM << "bpm " << i << ": " << bpm_pos[i] << endl;
		}
		bpm_size = mycatsync->gof2_bpm_idx.size();
#else
		try {
			bpm_crl = mycatsync->match_orb[thread_index]->read_attribute(bpm_attrnames[thread_index]);
			for (uint32_t r = 0; r < bpm_crl.size(); r++) {	
				try {
					bpm_crl[r] >> bpm_pos[r];
					bpm_pos[r] /= 1000; // convert mm into m
					
				}	
				catch (Tango::DevFailed &e) {
					TangoSys_MemStream error_stream;
					error_stream << ", bpm idx " << r << ", error " << e.errors[0].desc;
					INFO_STREAM << __FUNCTION__ << error_stream.str() << endl;
					strncpy(mycatsync->attr_MatchThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
					
					err_flag = true;
					bpm_pos[r] = NAN;	
				}
		 	}
			bpm_size = bpm_crl.size();
		}
		catch (Tango::DevFailed &e) {
				TangoSys_MemStream error_stream;
				error_stream << ", writing " << bpm_attrnames[thread_index] << 
					" error " << e.errors[0].desc;
				INFO_STREAM <<  __FUNCTION__ << ", " << error_stream.str() << endl;
				strncpy(mycatsync->attr_MatchThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);

		}
		catch(...) {
			TangoSys_MemStream error_stream;
			error_stream << ", writing " << bpm_attrnames[thread_index] << 
					" generic error";
			INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
			strncpy(mycatsync->attr_MatchThreadErrMsg_read[thread_index], error_stream.str().c_str(), CATSYNC_LEN);
		
		}
#endif

		// if there is an error goto the start of the loop
		if (err_flag) {
			omni_thread::sleep(1,0); continue;
		}

		try {
			vector <double> vval;
			vector <string> ss;

			for (uint32_t r = 0; r < bpm_size; r++) {
				vval.push_back(bpm_pos[r]);
			}

			Tango::AttributeInfoEx attr_info = mycatsync->catdevice->get_attribute_config(sim_attrnames[thread_index]);
			Din = insert_values(attr_info, sim_attrnames[thread_index], vval, ss);
			mycatsync->catdevice->write_attribute(*Din);
			if (Din) {
				delete Din; Din = 0;
			}
			(mycatsync->attr_MatchThreadCnt_read[thread_index])++;


		}
		catch (Tango::DevFailed &e) {
			TangoSys_MemStream error_stream;
			error_stream << e.errors[0].desc << ", error setting " << 
				mycatsync->cATDSDevice << ", " << 
				sim_attrnames[thread_index];
			INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
			if (Din) {
				delete Din; Din = 0;
			}
			(mycatsync->attr_MatchThreadErrCnt_read[thread_index])++;
			omni_thread::sleep(1,0); // 1 sec in case of errors
		}
		catch(...) {
			TangoSys_MemStream error_stream;
			error_stream << "generic error, error setting " << 
				mycatsync->cATDSDevice << ", " << 
				sim_attrnames[thread_index];
			INFO_STREAM << __FUNCTION__ << ", " << error_stream.str() << endl;
			if (Din) {
				delete Din; Din = 0;
			}
			(mycatsync->attr_MatchThreadErrCnt_read[thread_index])++;
			omni_thread::sleep(1,0); // 1 sec in case of errors	
		}

		// 
		if ((thread_index == 0) || ((mycatsync->tunehor_device != 0) && (mycatsync->tunever_device != 0))) {
			try {
				float tune_hor, tune_ver;

				mycatsync->tunehor_device->read_attribute(mycatsync->tunehor_attribute.c_str()) >> tune_hor;
				mycatsync->tunever_device->read_attribute(mycatsync->tunever_attribute.c_str()) >> tune_ver;			
		
				vector <double> tuneref;
				tuneref.push_back((double)tune_hor);
				tuneref.push_back((double)tune_ver);
				Tango::DeviceAttribute wr_attribute("tunefb_ref_sensors", tuneref);
				mycatsync->catdevice->write_attribute(wr_attribute);
			}
			catch(...) {}
		}




		omni_thread::sleep(0,10000000); // 10ms
		
	}

	mycatsync->abort_match_orb_flag[thread_index] = false;	
	mycatsync->run_match_orb_flag[thread_index] = false;
}

}

